# Copyright (C) 2015 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import sys
import time
import unittest
import inspect

import collections
collections.Callable = collections.abc.Callable

from lib.noseplugin import parser_option

from lib import base
from lib.base import (
    Bridge,
    BGP_FSM_ESTABLISHED,
    BGP_ATTR_TYPE_COMMUNITIES,
    BGP_ATTR_TYPE_EXTENDED_COMMUNITIES,
    local,
)
from lib.gobgp import GoBGPContainer
from lib.quagga import QuaggaBGPContainer
from lib.exabgp import ExaBGPContainer


counter = 1
_SCENARIOS = {}


def register_scenario(cls):
    global counter
    _SCENARIOS[counter] = cls
    counter += 1


def lookup_scenario(name):
    for value in list(_SCENARIOS.values()):
        if value.__name__ == name:
            return value
    return None


def wait_for(f, timeout=120):
    interval = 1
    count = 0
    while True:
        if f():
            return

        time.sleep(interval)
        count += interval
        if count >= timeout:
            raise Exception('timeout')


@register_scenario
class ImportPolicy(object):
    """
    No.1 import-policy test
                            --------------------------------
    e1 ->(192.168.2.0/24)-> | ->  q1-rib -> q1-adj-rib-out | --> q1
                            |                              |
                            | ->x q2-rib                   |
                            --------------------------------
    """
    @staticmethod
    def boot(env):
        gobgp_ctn_image_name = env.parser_option.gobgp_image
        log_level = env.parser_option.gobgp_log_level
        g1 = GoBGPContainer(name='g1', asn=65000, router_id='192.168.0.1',
                            ctn_image_name=gobgp_ctn_image_name,
                            log_level=log_level)
        e1 = ExaBGPContainer(name='e1', asn=65001, router_id='192.168.0.2')
        q1 = QuaggaBGPContainer(name='q1', asn=65002, router_id='192.168.0.3')
        q2 = QuaggaBGPContainer(name='q2', asn=65003, router_id='192.168.0.4')

        ctns = [g1, e1, q1, q2]
        initial_wait_time = max(ctn.run() for ctn in ctns)
        time.sleep(initial_wait_time)

        for q in [e1, q1, q2]:
            g1.add_peer(q, is_rs_client=True)
            q.add_peer(g1)

        env.g1 = g1
        env.e1 = e1
        env.q1 = q1
        env.q2 = q2

    @staticmethod
    def setup(env):
        g1 = env.g1
        e1 = env.e1
        q1 = env.q1
        q2 = env.q2

        p0 = {'ip-prefix': '192.168.0.0/16',
              'masklength-range': '16..24'}

        ps0 = {'prefix-set-name': 'ps0',
               'prefix-list': [p0]}
        g1.set_prefix_set(ps0)

        ns0 = {'neighbor-set-name': 'ns0',
               'neighbor-info-list': [g1.peers[q2]['neigh_addr'].split('/')[0]]}
        g1.set_neighbor_set(ns0)

        st0 = {'name': 'st0',
               'conditions': {
                   'match-prefix-set': {'prefix-set': ps0['prefix-set-name']},
                   'match-neighbor-set': {'neighbor-set': ns0['neighbor-set-name']}},
               'actions': {'route-disposition': 'reject-route'}}

        policy = {'name': 'policy0',
                  'statements': [st0]}
        g1.add_policy(policy, q2, 'export')

        # this will be blocked
        e1.add_route('192.168.2.0/24')
        # this will pass
        e1.add_route('192.168.2.0/15')

        for c in [e1, q1, q2]:
            g1.wait_for(BGP_FSM_ESTABLISHED, c)

    @staticmethod
    def check(env):
        wait_for(lambda: len(env.g1.get_local_rib(env.q1)) == 2)
        wait_for(lambda: len(env.g1.get_adj_rib_out(env.q1)) == 2)
        wait_for(lambda: len(env.q1.get_global_rib()) == 2)
        wait_for(lambda: len(env.g1.get_local_rib(env.q2)) == 2)
        wait_for(lambda: len(env.g1.get_adj_rib_out(env.q2)) == 1)
        wait_for(lambda: len(env.q2.get_global_rib()) == 1)

    @staticmethod
    def executor(env):
        lookup_scenario("ImportPolicy").boot(env)
        lookup_scenario("ImportPolicy").setup(env)
        lookup_scenario("ImportPolicy").check(env)


@register_scenario
class ExportPolicy(object):
    """
    No.2 export-policy test
                            --------------------------------
    e1 ->(192.168.2.0/24)-> | -> q1-rib ->  q1-adj-rib-out | --> q1
                            |                              |
                            | -> q2-rib ->x q2-adj-rib-out |
                            --------------------------------
    """
    @staticmethod
    def boot(env):
        lookup_scenario("ImportPolicy").boot(env)

    @staticmethod
    def setup(env):
        g1 = env.g1
        e1 = env.e1
        q1 = env.q1
        q2 = env.q2

        p0 = {'ip-prefix': '192.168.0.0/16',
              'masklength-range': '16..24'}

        ps0 = {'prefix-set-name': 'ps0',
               'prefix-list': [p0]}
        g1.set_prefix_set(ps0)

        ns0 = {'neighbor-set-name': 'ns0',
               'neighbor-info-list': [g1.peers[q2]['neigh_addr'].split('/')[0]]}
        g1.set_neighbor_set(ns0)

        st0 = {'name': 'st0',
               'conditions': {
                   'match-prefix-set': {'prefix-set': ps0['prefix-set-name']},
                   'match-neighbor-set': {'neighbor-set': ns0['neighbor-set-name']}},
               'actions': {'route-disposition': 'reject-route'}}

        policy = {'name': 'policy0',
                  'statements': [st0]}
        g1.add_policy(policy, q2, 'export')

        # this will be blocked
        e1.add_route('192.168.2.0/24')
        # this will pass
        e1.add_route('192.168.2.0/15')

        for c in [e1, q1, q2]:
            g1.wait_for(BGP_FSM_ESTABLISHED, c)

    @staticmethod
    def check(env):
        g1 = env.g1
        q1 = env.q1
        q2 = env.q2
        wait_for(lambda: len(g1.get_local_rib(q1)) == 2)
        wait_for(lambda: len(g1.get_adj_rib_out(q1)) == 2)
        wait_for(lambda: len(q1.get_global_rib()) == 2)
        wait_for(lambda: len(g1.get_local_rib(q2)) == 2)
        wait_for(lambda: len(g1.get_adj_rib_out(q2)) == 1)
        wait_for(lambda: len(q2.get_global_rib()) == 1)

    @staticmethod
    def executor(env):
        lookup_scenario("ExportPolicy").boot(env)
        lookup_scenario("ExportPolicy").setup(env)
        lookup_scenario("ExportPolicy").check(env)


@register_scenario
class ImportPolicyUpdate(object):
    """
    No.3 import-policy update test

    r1:192.168.2.0/24
    r2:192.168.20.0/24
    r3:192.168.200.0/24
                      -------------------------------------------------
                      | q1                                            |
    e1 ->(r1,r2,r3)-> | ->(r1,r2,r3)-> rib ->(r1,r2,r3)-> adj-rib-out | ->(r1,r2,r3)-> q1
                      |                                               |
                      | q2                                            |
                      | ->(r1)->       rib ->(r1)->       adj-rib-out | ->(r1)-> q2
                      -------------------------------------------------
                 |
         update gobgp.conf
                 |
                 V
                      -------------------------------------------------
                      | q1                                            |
    e1 ->(r1,r2,r3)-> | ->(r1,r2,r3)-> rib ->(r1,r2,r3)-> adj-rib-out | ->(r1,r2,r3)-> q1
                      |                                               |
                      | q2                                            |
                      | ->(r1,r3)->    rib ->(r1,r3)->    adj-rib-out | ->(r1,r3)-> q2
                      -------------------------------------------------
    """
    @staticmethod
    def boot(env):
        lookup_scenario("ImportPolicy").boot(env)

    @staticmethod
    def setup(env):
        g1 = env.g1
        e1 = env.e1
        q1 = env.q1
        q2 = env.q2

        p0 = {'ip-prefix': '192.168.20.0/24'}
        p1 = {'ip-prefix': '192.168.200.0/24'}

        ps0 = {'prefix-set-name': 'ps0',
               'prefix-list': [p0, p1]}
        g1.set_prefix_set(ps0)

        ns0 = {'neighbor-set-name': 'ns0',
               'neighbor-info-list': [g1.peers[q2]['neigh_addr'].split('/')[0]]}
        g1.set_neighbor_set(ns0)

        st0 = {'name': 'st0',
               'conditions': {
                   'match-prefix-set': {'prefix-set': ps0['prefix-set-name']},
                   'match-neighbor-set': {'neighbor-set': ns0['neighbor-set-name']}},
               'actions': {'route-disposition': 'reject-route'}}

        policy = {'name': 'policy0',
                  'statements': [st0]}
        g1.add_policy(policy, q2, 'export')

        e1.add_route('192.168.2.0/24')
        e1.add_route('192.168.20.0/24')
        e1.add_route('192.168.200.0/24')

        for c in [e1, q1, q2]:
            g1.wait_for(BGP_FSM_ESTABLISHED, c)

    @staticmethod
    def check(env):
        g1 = env.g1
        # e1 = env.e1
        q1 = env.q1
        q2 = env.q2
        wait_for(lambda: len(g1.get_local_rib(q1)) == 3)
        wait_for(lambda: len(g1.get_adj_rib_out(q1)) == 3)
        wait_for(lambda: len(q1.get_global_rib()) == 3)
        wait_for(lambda: len(g1.get_local_rib(q2)) == 3)
        wait_for(lambda: len(g1.get_adj_rib_out(q2)) == 1)
        wait_for(lambda: len(q2.get_global_rib()) == 1)

    @staticmethod
    def setup2(env):
        g1 = env.g1
        e1 = env.e1
        q1 = env.q1
        q2 = env.q2
        g1.clear_policy()

        p0 = {'ip-prefix': '192.168.20.0/24'}

        ps0 = {'prefix-set-name': 'ps0',
               'prefix-list': [p0]}
        g1.set_prefix_set(ps0)

        ns0 = {'neighbor-set-name': 'ns0',
               'neighbor-info-list': [g1.peers[q2]['neigh_addr'].split('/')[0]]}
        g1.set_neighbor_set(ns0)

        st0 = {'name': 'st0',
               'conditions': {
                   'match-prefix-set': {'prefix-set': ps0['prefix-set-name']},
                   'match-neighbor-set': {'neighbor-set': ns0['neighbor-set-name']}},
               'actions': {'route-disposition': 'reject-route'}}

        policy = {'name': 'policy0',
                  'statements': [st0]}
        g1.add_policy(policy, q2, 'export')
        g1.reset(e1)

        for c in [e1, q1, q2]:
            g1.wait_for(BGP_FSM_ESTABLISHED, c)

    @staticmethod
    def check2(env):
        g1 = env.g1
        # e1 = env.e1
        q1 = env.q1
        q2 = env.q2
        wait_for(lambda: len(g1.get_local_rib(q1)) == 3)
        wait_for(lambda: len(g1.get_adj_rib_out(q1)) == 3)
        wait_for(lambda: len(q1.get_global_rib()) == 3)
        wait_for(lambda: len(g1.get_local_rib(q2)) == 3)
        wait_for(lambda: len(g1.get_adj_rib_out(q2)) == 2)
        wait_for(lambda: len(q2.get_global_rib()) == 2)

    @staticmethod
    def executor(env):
        lookup_scenario("ImportPolicyUpdate").boot(env)
        lookup_scenario("ImportPolicyUpdate").setup(env)
        lookup_scenario("ImportPolicyUpdate").check(env)
        lookup_scenario("ImportPolicyUpdate").setup2(env)
        lookup_scenario("ImportPolicyUpdate").check2(env)


@register_scenario
class ExportPolicyUpdate(object):
    """
    No.4 export-policy update test

    r1:192.168.2.0
    r2:192.168.20.0
    r3:192.168.200.0
                      -------------------------------------------------
                      | q1                                            |
    e1 ->(r1,r2,r3)-> | ->(r1,r2,r3)-> rib ->(r1,r2,r3)-> adj-rib-out | ->(r1,r2,r3)-> q1
                      |                                               |
                      | q2                                            |
                      | ->(r1,r2,r3)-> rib ->(r1)->       adj-rib-out | ->(r1)-> q2
                      -------------------------------------------------
                 |
         update gobgp.conf
                 |
                 V
                      -------------------------------------------------
                      | q1                                            |
    e1 ->(r1,r2,r3)-> | ->(r1,r2,r3)-> rib ->(r1,r2,r3)-> adj-rib-out | ->(r1,r2,r3)-> q1
                      |                                               |
                      | q2                                            |
                      | ->(r1,r2,r3)-> rib ->(r1,r3)->    adj-rib-out | ->(r1,r3)-> q2
                      -------------------------------------------------
    """
    @staticmethod
    def boot(env):
        lookup_scenario("ImportPolicy").boot(env)

    @staticmethod
    def setup(env):
        g1 = env.g1
        e1 = env.e1
        q1 = env.q1
        q2 = env.q2
        p0 = {'ip-prefix': '192.168.20.0/24'}
        p1 = {'ip-prefix': '192.168.200.0/24'}

        ps0 = {'prefix-set-name': 'ps0',
               'prefix-list': [p0, p1]}
        g1.set_prefix_set(ps0)

        ns0 = {'neighbor-set-name': 'ns0',
               'neighbor-info-list': [g1.peers[q2]['neigh_addr'].split('/')[0]]}
        g1.set_neighbor_set(ns0)

        st0 = {'name': 'st0',
               'conditions': {
                   'match-prefix-set': {'prefix-set': ps0['prefix-set-name']},
                   'match-neighbor-set': {'neighbor-set': ns0['neighbor-set-name']}},
               'actions': {'route-disposition': 'reject-route'}}

        policy = {'name': 'policy0',
                  'statements': [st0]}
        g1.add_policy(policy, q2, 'export')

        e1.add_route('192.168.2.0/24')
        e1.add_route('192.168.20.0/24')
        e1.add_route('192.168.200.0/24')

        for c in [e1, q1, q2]:
            g1.wait_for(BGP_FSM_ESTABLISHED, c)

    @staticmethod
    def check(env):
        g1 = env.g1
        # e1 = env.e1
        q1 = env.q1
        q2 = env.q2
        wait_for(lambda: len(g1.get_local_rib(q1)) == 3)
        wait_for(lambda: len(g1.get_adj_rib_out(q1)) == 3)
        wait_for(lambda: len(q1.get_global_rib()) == 3)
        wait_for(lambda: len(g1.get_local_rib(q2)) == 3)
        wait_for(lambda: len(g1.get_adj_rib_out(q2)) == 1)
        wait_for(lambda: len(q2.get_global_rib()) == 1)

    @staticmethod
    def setup2(env):
        g1 = env.g1
        e1 = env.e1
        q1 = env.q1
        q2 = env.q2
        g1.clear_policy()

        p0 = {'ip-prefix': '192.168.20.0/24'}

        ps0 = {'prefix-set-name': 'ps0',
               'prefix-list': [p0]}
        g1.set_prefix_set(ps0)

        ns0 = {'neighbor-set-name': 'ns0',
               'neighbor-info-list': [g1.peers[q2]['neigh_addr'].split('/')[0]]}
        g1.set_neighbor_set(ns0)

        st0 = {'name': 'st0',
               'conditions': {
                   'match-prefix-set': {'prefix-set': ps0['prefix-set-name']},
                   'match-neighbor-set': {'neighbor-set': ns0['neighbor-set-name']}},
               'actions': {'route-disposition': 'reject-route'}}

        policy = {'name': 'policy0',
                  'statements': [st0]}
        g1.add_policy(policy, q2, 'export')

        # we need hard reset to flush q2's local rib
        g1.reset(e1)

        for c in [e1, q1, q2]:
            g1.wait_for(BGP_FSM_ESTABLISHED, c)

    @staticmethod
    def check2(env):
        g1 = env.g1
        # e1 = env.e1
        q1 = env.q1
        q2 = env.q2
        wait_for(lambda: len(g1.get_local_rib(q1)) == 3)
        wait_for(lambda: len(g1.get_adj_rib_out(q1)) == 3)
        wait_for(lambda: len(q1.get_global_rib()) == 3)
        wait_for(lambda: len(g1.get_local_rib(q2)) == 3)
        wait_for(lambda: len(g1.get_adj_rib_out(q2)) == 2)
        wait_for(lambda: len(q2.get_global_rib()) == 2)

    @staticmethod
    def executor(env):
        lookup_scenario("ExportPolicyUpdate").boot(env)
        lookup_scenario("ExportPolicyUpdate").setup(env)
        lookup_scenario("ExportPolicyUpdate").check(env)
        lookup_scenario("ExportPolicyUpdate").setup2(env)
        lookup_scenario("ExportPolicyUpdate").check2(env)


@register_scenario
class ImportPolicyIPV6(object):
    """
    No.5 IPv6 import-policy test

    r1=2001::/64
    r2=2001::/63
                   -------------------------------------------------
    e1 ->(r1,r2)-> | ->(r1,r2)-> q1-rib ->(r1,r2)-> q1-adj-rib-out | ->(r1,r2)-> q1
                   |                                               |
                   | ->(r2)   -> q2-rib ->(r2)   -> q2-adj-rib-out | ->(r2)-> q2
                   -------------------------------------------------
    """
    @staticmethod
    def boot(env):
        gobgp_ctn_image_name = env.parser_option.gobgp_image
        log_level = env.parser_option.gobgp_log_level
        g1 = GoBGPContainer(name='g1', asn=65000, router_id='192.168.0.1',
                            ctn_image_name=gobgp_ctn_image_name,
                            log_level=log_level)
        e1 = ExaBGPContainer(name='e1', asn=65001, router_id='192.168.0.2')
        q1 = QuaggaBGPContainer(name='q1', asn=65002, router_id='192.168.0.3')
        q2 = QuaggaBGPContainer(name='q2', asn=65003, router_id='192.168.0.4')

        ctns = [g1, e1, q1, q2]
        initial_wait_time = max(ctn.run() for ctn in ctns)
        time.sleep(initial_wait_time)

        br01 = Bridge(name='br01', subnet='2001::/96')
        [br01.addif(ctn) for ctn in ctns]

        for q in [e1, q1, q2]:
            g1.add_peer(q, is_rs_client=True, bridge=br01.name, v6=True)
            q.add_peer(g1, bridge=br01.name, v6=True)

            env.g1 = g1
            env.e1 = e1
            env.q1 = q1
            env.q2 = q2

    @staticmethod
    def setup(env):
        g1 = env.g1
        e1 = env.e1
        q1 = env.q1
        q2 = env.q2

        p0 = {'ip-prefix': '2001::/32',
              'masklength-range': '64..128'}

        ps0 = {'prefix-set-name': 'ps0',
               'prefix-list': [p0]}
        g1.set_prefix_set(ps0)

        ns0 = {'neighbor-set-name': 'ns0',
               'neighbor-info-list': [g1.peers[q2]['neigh_addr'].split('/')[0]]}
        g1.set_neighbor_set(ns0)

        st0 = {'name': 'st0',
               'conditions': {
                   'match-prefix-set': {'prefix-set': ps0['prefix-set-name']},
                   'match-neighbor-set': {'neighbor-set': ns0['neighbor-set-name']}},
               'actions': {'route-disposition': 'reject-route'}}

        policy = {'name': 'policy0',
                  'statements': [st0]}
        g1.add_policy(policy, q2, 'export')

        # this will be blocked
        e1.add_route('2001::/64', rf='ipv6')
        # this will pass
        e1.add_route('2001::/63', rf='ipv6')

        for c in [e1, q1, q2]:
            g1.wait_for(BGP_FSM_ESTABLISHED, c)

    @staticmethod
    def check(env):
        wait_for(lambda: len(env.g1.get_local_rib(env.q1, rf='ipv6')) == 2)
        wait_for(lambda: len(env.g1.get_adj_rib_out(env.q1, rf='ipv6')) == 2)
        wait_for(lambda: len(env.q1.get_global_rib(rf='ipv6')) == 2)
        wait_for(lambda: len(env.g1.get_local_rib(env.q2, rf='ipv6')) == 2)
        wait_for(lambda: len(env.g1.get_adj_rib_out(env.q2, rf='ipv6')) == 1)
        wait_for(lambda: len(env.q2.get_global_rib(rf='ipv6')) == 1)

    @staticmethod
    def executor(env):
        lookup_scenario("ImportPolicyIPV6").boot(env)
        lookup_scenario("ImportPolicyIPV6").setup(env)
        lookup_scenario("ImportPolicyIPV6").check(env)


@register_scenario
class ExportPolicyIPV6(object):
    """
    No.6 IPv6 export-policy test

    r1=2001::/64
    r2=2001::/63
                   -------------------------------------------------
    e1 ->(r1,r2)-> | ->(r1,r2)-> q1-rib ->(r1,r2)-> q1-adj-rib-out | ->(r1,r2)-> q1
                   |                                               |
                   | ->(r1,r2)-> q2-rib ->(r2)   -> q2-adj-rib-out | ->(r2)-> q2
                   -------------------------------------------------
    """
    @staticmethod
    def boot(env):
        lookup_scenario('ImportPolicyIPV6').boot(env)

    @staticmethod
    def setup(env):
        g1 = env.g1
        e1 = env.e1
        q1 = env.q1
        q2 = env.q2

        p0 = {'ip-prefix': '2001::/32',
              'masklength-range': '64..128'}

        ps0 = {'prefix-set-name': 'ps0',
               'prefix-list': [p0]}
        g1.set_prefix_set(ps0)

        ns0 = {'neighbor-set-name': 'ns0',
               'neighbor-info-list': [g1.peers[q2]['neigh_addr'].split('/')[0]]}
        g1.set_neighbor_set(ns0)

        st0 = {'name': 'st0',
               'conditions': {
                   'match-prefix-set': {'prefix-set': ps0['prefix-set-name']},
                   'match-neighbor-set': {'neighbor-set': ns0['neighbor-set-name']}},
               'actions': {'route-disposition': 'reject-route'}}

        policy = {'name': 'policy0',
                  'statements': [st0]}
        g1.add_policy(policy, q2, 'export')

        # this will be blocked
        e1.add_route('2001::/64', rf='ipv6')
        # this will pass
        e1.add_route('2001::/63', rf='ipv6')

        for c in [e1, q1, q2]:
            g1.wait_for(BGP_FSM_ESTABLISHED, c)

    @staticmethod
    def check(env):
        wait_for(lambda: len(env.g1.get_local_rib(env.q1, rf='ipv6')) == 2)
        wait_for(lambda: len(env.g1.get_adj_rib_out(env.q1, rf='ipv6')) == 2)
        wait_for(lambda: len(env.q1.get_global_rib(rf='ipv6')) == 2)
        wait_for(lambda: len(env.g1.get_local_rib(env.q2, rf='ipv6')) == 2)
        wait_for(lambda: len(env.g1.get_adj_rib_out(env.q2, rf='ipv6')) == 1)
        wait_for(lambda: len(env.q2.get_global_rib(rf='ipv6')) == 1)

    @staticmethod
    def executor(env):
        lookup_scenario("ExportPolicyIPV6").boot(env)
        lookup_scenario("ExportPolicyIPV6").setup(env)
        lookup_scenario("ExportPolicyIPV6").check(env)


@register_scenario
class ImportPolicyIPV6Update(object):
    """
    No.7 IPv6 import-policy update test
    r1=2001:0:10:2::/64
    r2=2001:0:10:20::/64
    r3=2001:0:10:200::/64
                      -------------------------------------------------
                      | q1                                            |
    e1 ->(r1,r2,r3)-> | ->(r1,r2,r3)-> rib ->(r1,r2,r3)-> adj-rib-out | ->(r1,r2,r3)-> q1
                      |                                               |
                      | q2                                            |
                      | ->(r1)->       rib ->(r1)->       adj-rib-out | ->(r1)-> q2
                      -------------------------------------------------
                 |
         update gobgp.conf
                 |
                 V
                      -------------------------------------------------
                      | q1                                            |
    e1 ->(r1,r2,r3)-> | ->(r1,r2,r3)-> rib ->(r1,r2,r3)-> adj-rib-out | ->(r1,r2,r3)-> q1
                      |                                               |
                      | q2                                            |
                      | ->(r1,r3)->    rib ->(r1,r3)->    adj-rib-out | ->(r1,r3)-> q2
                      -------------------------------------------------
    """
    @staticmethod
    def boot(env):
        lookup_scenario('ImportPolicyIPV6').boot(env)

    @staticmethod
    def setup(env):
        g1 = env.g1
        e1 = env.e1
        q1 = env.q1
        q2 = env.q2

        p0 = {'ip-prefix': '2001:0:10:2::/64'}
        p1 = {'ip-prefix': '2001:0:10:20::/64'}

        ps0 = {'prefix-set-name': 'ps0',
               'prefix-list': [p0, p1]}
        g1.set_prefix_set(ps0)

        ns0 = {'neighbor-set-name': 'ns0',
               'neighbor-info-list': [g1.peers[q2]['neigh_addr'].split('/')[0]]}
        g1.set_neighbor_set(ns0)

        st0 = {'name': 'st0',
               'conditions': {
                   'match-prefix-set': {'prefix-set': ps0['prefix-set-name']},
                   'match-neighbor-set': {'neighbor-set': ns0['neighbor-set-name']}},
               'actions': {'route-disposition': 'reject-route'}}

        policy = {'name': 'policy0',
                  'statements': [st0]}
        g1.add_policy(policy, q2, 'export')

        e1.add_route('2001:0:10:2::/64', rf='ipv6')
        e1.add_route('2001:0:10:20::/64', rf='ipv6')
        e1.add_route('2001:0:10:200::/64', rf='ipv6')

        for c in [e1, q1, q2]:
            g1.wait_for(BGP_FSM_ESTABLISHED, c)

    @staticmethod
    def check(env):
        wait_for(lambda: len(env.g1.get_local_rib(env.q1, rf='ipv6')) == 3)
        wait_for(lambda: len(env.g1.get_adj_rib_out(env.q1, rf='ipv6')) == 3)
        wait_for(lambda: len(env.q1.get_global_rib(rf='ipv6')) == 3)
        wait_for(lambda: len(env.g1.get_local_rib(env.q2, rf='ipv6')) == 3)
        wait_for(lambda: len(env.g1.get_adj_rib_out(env.q2, rf='ipv6')) == 1)
        wait_for(lambda: len(env.q2.get_global_rib(rf='ipv6')) == 1)

    @staticmethod
    def setup2(env):
        g1 = env.g1
        e1 = env.e1
        q1 = env.q1
        q2 = env.q2

        p0 = {'ip-prefix': '2001:0:10:2::/64'}

        ps0 = {'prefix-set-name': 'ps0',
               'prefix-list': [p0]}
        g1.set_prefix_set(ps0)

        ns0 = {'neighbor-set-name': 'ns0',
               'neighbor-info-list': [g1.peers[q2]['neigh_addr'].split('/')[0]]}
        g1.set_neighbor_set(ns0)

        st0 = {'name': 'st0',
               'conditions': {
                   'match-prefix-set': {'prefix-set': ps0['prefix-set-name']},
                   'match-neighbor-set': {'neighbor-set': ns0['neighbor-set-name']}},
               'actions': {
                   'route-disposition': 'reject-route'}}

        policy = {'name': 'policy0',
                  'statements': [st0]}
        g1.add_policy(policy, q2, 'export')
        g1.reset(e1)

        for c in [e1, q1, q2]:
            g1.wait_for(BGP_FSM_ESTABLISHED, c)

    @staticmethod
    def check2(env):
        wait_for(lambda: len(env.g1.get_local_rib(env.q1, rf='ipv6')) == 3)
        wait_for(lambda: len(env.g1.get_adj_rib_out(env.q1, rf='ipv6')) == 3)
        wait_for(lambda: len(env.q1.get_global_rib(rf='ipv6')) == 3)
        wait_for(lambda: len(env.g1.get_local_rib(env.q2, rf='ipv6')) == 3)
        wait_for(lambda: len(env.g1.get_adj_rib_out(env.q2, rf='ipv6')) == 2)
        wait_for(lambda: len(env.q2.get_global_rib(rf='ipv6')) == 2)

    @staticmethod
    def executor(env):
        lookup_scenario("ImportPolicyIPV6Update").boot(env)
        lookup_scenario("ImportPolicyIPV6Update").setup(env)
        lookup_scenario("ImportPolicyIPV6Update").check(env)
        lookup_scenario("ImportPolicyIPV6Update").setup2(env)
        lookup_scenario("ImportPolicyIPV6Update").check2(env)


@register_scenario
class ExportPolicyIPv6Update(object):
    """
    No.8 IPv6 export-policy update test
    r1=2001:0:10:2::/64
    r2=2001:0:10:20::/64
    r3=2001:0:10:200::/64
                      -------------------------------------------------
                      | q1                                            |
    e1 ->(r1,r2,r3)-> | ->(r1,r2,r3)-> rib ->(r1,r2,r3)-> adj-rib-out | ->(r1,r2,r3)-> q1
                      |                                               |
                      | q2                                            |
                      | ->(r1,r2,r3)-> rib ->(r1)->       adj-rib-out | ->(r1)-> q2
                      -------------------------------------------------
                 |
         update gobgp.conf
                 |
                 V
                      -------------------------------------------------
                      | q1                                            |
    e1 ->(r1,r2,r3)-> | ->(r1,r2,r3)-> rib ->(r1,r2,r3)-> adj-rib-out | ->(r1,r2,r3)-> q1
                      |                                               |
                      | q2                                            |
                      | ->(r1,r2,r3)-> rib ->(r1,r3)->    adj-rib-out | ->(r1,r3)-> q2
                      -------------------------------------------------
    """
    @staticmethod
    def boot(env):
        lookup_scenario('ImportPolicyIPV6').boot(env)

    @staticmethod
    def setup(env):
        g1 = env.g1
        e1 = env.e1
        q1 = env.q1
        q2 = env.q2

        p0 = {'ip-prefix': '2001:0:10:2::/64'}
        p1 = {'ip-prefix': '2001:0:10:20::/64'}

        ps0 = {'prefix-set-name': 'ps0',
               'prefix-list': [p0, p1]}
        g1.set_prefix_set(ps0)

        ns0 = {'neighbor-set-name': 'ns0',
               'neighbor-info-list': [g1.peers[q2]['neigh_addr'].split('/')[0]]}
        g1.set_neighbor_set(ns0)

        st0 = {'name': 'st0',
               'conditions': {
                   'match-prefix-set': {'prefix-set': ps0['prefix-set-name']},
                   'match-neighbor-set': {'neighbor-set': ns0['neighbor-set-name']}},
               'actions': {'route-disposition': 'reject-route'}}

        policy = {'name': 'policy0',
                  'statements': [st0]}
        g1.add_policy(policy, q2, 'export')

        e1.add_route('2001:0:10:2::/64', rf='ipv6')
        e1.add_route('2001:0:10:20::/64', rf='ipv6')
        e1.add_route('2001:0:10:200::/64', rf='ipv6')

        for c in [e1, q1, q2]:
            g1.wait_for(BGP_FSM_ESTABLISHED, c)

    @staticmethod
    def check(env):
        wait_for(lambda: len(env.g1.get_local_rib(env.q1, rf='ipv6')) == 3)
        wait_for(lambda: len(env.g1.get_adj_rib_out(env.q1, rf='ipv6')) == 3)
        wait_for(lambda: len(env.q1.get_global_rib(rf='ipv6')) == 3)
        wait_for(lambda: len(env.g1.get_local_rib(env.q2, rf='ipv6')) == 3)
        wait_for(lambda: len(env.g1.get_adj_rib_out(env.q2, rf='ipv6')) == 1)
        wait_for(lambda: len(env.q2.get_global_rib(rf='ipv6')) == 1)

    @staticmethod
    def setup2(env):
        g1 = env.g1
        e1 = env.e1
        q1 = env.q1
        q2 = env.q2

        p0 = {'ip-prefix': '2001:0:10:2::/64'}

        ps0 = {'prefix-set-name': 'ps0',
               'prefix-list': [p0]}
        g1.set_prefix_set(ps0)

        ns0 = {'neighbor-set-name': 'ns0',
               'neighbor-info-list': [g1.peers[q2]['neigh_addr'].split('/')[0]]}
        g1.set_neighbor_set(ns0)

        st0 = {'name': 'st0',
               'conditions': {
                   'match-prefix-set': {'prefix-set': ps0['prefix-set-name']},
                   'match-neighbor-set': {'neighbor-set': ns0['neighbor-set-name']}},
               'actions': {'route-disposition': 'reject-route'}}

        policy = {'name': 'policy0',
                  'statements': [st0]}
        g1.add_policy(policy, q2, 'export')
        g1.reset(e1)

        for c in [e1, q1, q2]:
            g1.wait_for(BGP_FSM_ESTABLISHED, c)

    @staticmethod
    def check2(env):
        wait_for(lambda: len(env.g1.get_local_rib(env.q1, rf='ipv6')) == 3)
        wait_for(lambda: len(env.g1.get_adj_rib_out(env.q1, rf='ipv6')) == 3)
        wait_for(lambda: len(env.q1.get_global_rib(rf='ipv6')) == 3)
        wait_for(lambda: len(env.g1.get_local_rib(env.q2, rf='ipv6')) == 3)
        wait_for(lambda: len(env.g1.get_adj_rib_out(env.q2, rf='ipv6')) == 2)
        wait_for(lambda: len(env.q2.get_global_rib(rf='ipv6')) == 2)

    @staticmethod
    def executor(env):
        lookup_scenario("ExportPolicyIPv6Update").boot(env)
        lookup_scenario("ExportPolicyIPv6Update").setup(env)
        lookup_scenario("ExportPolicyIPv6Update").check(env)
        lookup_scenario("ExportPolicyIPv6Update").setup2(env)
        lookup_scenario("ExportPolicyIPv6Update").check2(env)


@register_scenario
class ImportPolicyAsPathLengthCondition(object):
    """
    No.9 aspath length condition import-policy test
                              --------------------------------
    e1 ->(aspath_length=10)-> | ->  q1-rib -> q1-adj-rib-out | --> q1
                              |                              |
                              | ->x q2-rib                   |
                              --------------------------------
    """
    @staticmethod
    def boot(env):
        lookup_scenario("ImportPolicy").boot(env)

    @staticmethod
    def setup(env):
        g1 = env.g1
        e1 = env.e1
        q1 = env.q1
        q2 = env.q2
        st0 = {'name': 'st0',
               'conditions': {'bgp-conditions': {'as-path-length': {'operator': 'ge',
                                                                    'value': 10}}},
               'actions': {'route-disposition': 'reject-route'}}

        policy = {'name': 'policy0',
                  'statements': [st0]}
        g1.add_policy(policy, q2, 'export')

        # this will be blocked (length 10, >= 10)
        e1.add_route('192.168.100.0/24', aspath=[e1.asn] + list(range(65100, 65109)))
        # this will pass (length 8, < 10)
        e1.add_route('192.168.200.0/24', aspath=[e1.asn] + list(range(65100, 65107)))

        for c in [e1, q1, q2]:
            g1.wait_for(BGP_FSM_ESTABLISHED, c)

    @staticmethod
    def check(env):
        g1 = env.g1
        # e1 = env.e1
        q1 = env.q1
        q2 = env.q2
        wait_for(lambda: len(g1.get_local_rib(q1)) == 2)
        wait_for(lambda: len(g1.get_adj_rib_out(q1)) == 2)
        wait_for(lambda: len(q1.get_global_rib()) == 2)
        wait_for(lambda: len(g1.get_local_rib(q2)) == 2)
        wait_for(lambda: len(g1.get_adj_rib_out(q2)) == 1)
        wait_for(lambda: len(q2.get_global_rib()) == 1)

    @staticmethod
    def executor(env):
        lookup_scenario("ImportPolicyAsPathLengthCondition").boot(env)
        lookup_scenario("ImportPolicyAsPathLengthCondition").setup(env)
        lookup_scenario("ImportPolicyAsPathLengthCondition").check(env)


@register_scenario
class ImportPolicyAsPathCondition(object):
    """
    No.10 aspath from condition import-policy test
                                --------------------------------
    e1 ->(aspath=[65100,...])-> | ->  q1-rib -> q1-adj-rib-out | --> q1
                                |                              |
                                | ->x q2-rib                   |
                                --------------------------------
    """
    @staticmethod
    def boot(env):
        lookup_scenario("ImportPolicy").boot(env)

    @staticmethod
    def setup(env):
        g1 = env.g1
        e1 = env.e1
        q1 = env.q1
        q2 = env.q2
        as0 = {'as-path-sets': [{'as-path-set-name': 'as0', 'as-path-list': ['^{0}'.format(e1.asn)]}]}

        g1.set_bgp_defined_set(as0)

        st0 = {'name': 'st0',
               'conditions': {'bgp-conditions': {'match-as-path-set': {'as-path-set': 'as0'}}},
               'actions': {'route-disposition': 'reject-route'}}

        policy = {'name': 'policy0',
                  'statements': [st0]}
        g1.add_policy(policy, q2, 'export')

        # this will be blocked (starts with e1.asn, matches ^65001)
        e1.add_route('192.168.100.0/24', aspath=[e1.asn] + list(range(65100, 65109)))
        # this will pass (does not start with e1.asn)
        e1.add_route('192.168.200.0/24', aspath=list(range(65100, 65109)))

        for c in [e1, q1, q2]:
            g1.wait_for(BGP_FSM_ESTABLISHED, c)

    @staticmethod
    def check(env):
        # same check function as previous No.1 scenario
        lookup_scenario("ImportPolicy").check(env)

    @staticmethod
    def executor(env):
        lookup_scenario("ImportPolicyAsPathCondition").boot(env)
        lookup_scenario("ImportPolicyAsPathCondition").setup(env)
        lookup_scenario("ImportPolicyAsPathCondition").check(env)


@register_scenario
class ImportPolicyAsPathAnyCondition(object):
    """
    No.11 aspath any condition import-policy test
                                   --------------------------------
    e1 ->(aspath=[...65098,...])-> | ->  q1-rib -> q1-adj-rib-out | --> q1
                                   |                              |
                                   | ->x q2-rib                   |
                                   --------------------------------
    """
    @staticmethod
    def boot(env):
        lookup_scenario("ImportPolicy").boot(env)

    @staticmethod
    def setup(env):
        g1 = env.g1
        e1 = env.e1
        q1 = env.q1
        q2 = env.q2
        as0 = {'as-path-sets': [{'as-path-set-name': 'as0', 'as-path-list': ['65098']}]}

        g1.set_bgp_defined_set(as0)

        st0 = {'name': 'st0',
               'conditions': {'bgp-conditions': {'match-as-path-set': {'as-path-set': 'as0'}}},
               'actions': {'route-disposition': 'reject-route'}}

        policy = {'name': 'policy0',
                  'statements': [st0]}
        g1.add_policy(policy, q2, 'export')

        # this will be blocked (contains 65098)
        e1.add_route('192.168.100.0/24', aspath=[65001, 65098, 65010])
        # this will pass (does not contain 65098)
        e1.add_route('192.168.200.0/24', aspath=[65001, 65100, 65010])

        for c in [e1, q1, q2]:
            g1.wait_for(BGP_FSM_ESTABLISHED, c)

    @staticmethod
    def check(env):
        # same check function as previous No.1 scenario
        lookup_scenario("ImportPolicy").check(env)

    @staticmethod
    def executor(env):
        lookup_scenario("ImportPolicyAsPathAnyCondition").boot(env)
        lookup_scenario("ImportPolicyAsPathAnyCondition").setup(env)
        lookup_scenario("ImportPolicyAsPathAnyCondition").check(env)


@register_scenario
class ImportPolicyAsPathOriginCondition(object):
    """
    No.12 aspath origin condition import-policy test
                                --------------------------------
    e1 ->(aspath=[...,65090])-> | ->  q1-rib -> q1-adj-rib-out | --> q1
                                |                              |
                                | ->x q2-rib                   |
                                --------------------------------
    """
    @staticmethod
    def boot(env):
        lookup_scenario("ImportPolicy").boot(env)

    @staticmethod
    def setup(env):
        g1 = env.g1
        e1 = env.e1
        q1 = env.q1
        q2 = env.q2
        as0 = {'as-path-sets': [{'as-path-set-name': 'as0', 'as-path-list': ['65090$']}]}

        g1.set_bgp_defined_set(as0)

        st0 = {'name': 'st0',
               'conditions': {'bgp-conditions': {'match-as-path-set': {'as-path-set': 'as0'}}},
               'actions': {'route-disposition': 'reject-route'}}

        policy = {'name': 'policy0',
                  'statements': [st0]}
        g1.add_policy(policy, q2, 'export')

        # this will be blocked (ends with 65090, matches 65090$)
        e1.add_route('192.168.100.0/24', aspath=[65001, 65098, 65090])
        # this will pass (does not end with 65090)
        e1.add_route('192.168.200.0/24', aspath=[65001, 65100, 65010])

        for c in [e1, q1, q2]:
            g1.wait_for(BGP_FSM_ESTABLISHED, c)

    @staticmethod
    def check(env):
        # same check function as previous No.1 scenario
        lookup_scenario("ImportPolicy").check(env)

    @staticmethod
    def executor(env):
        lookup_scenario("ImportPolicyAsPathOriginCondition").boot(env)
        lookup_scenario("ImportPolicyAsPathOriginCondition").setup(env)
        lookup_scenario("ImportPolicyAsPathOriginCondition").check(env)


@register_scenario
class ImportPolicyAsPathOnlyCondition(object):
    """
    No.13 aspath only condition import-policy test
                              --------------------------------
    e1 -> (aspath=[65100]) -> | ->  q1-rib -> q1-adj-rib-out | --> q1
                              |                              |
                              | ->x q2-rib                   |
                              --------------------------------
    """
    @staticmethod
    def boot(env):
        lookup_scenario("ImportPolicy").boot(env)

    @staticmethod
    def setup(env):
        g1 = env.g1
        e1 = env.e1
        q1 = env.q1
        q2 = env.q2
        as0 = {'as-path-sets': [{'as-path-set-name': 'as0', 'as-path-list': ['^65100$']}]}

        g1.set_bgp_defined_set(as0)

        st0 = {'name': 'st0',
               'conditions': {'bgp-conditions': {'match-as-path-set': {'as-path-set': 'as0'}}},
               'actions': {'route-disposition': 'reject-route'}}

        policy = {'name': 'policy0',
                  'statements': [st0]}
        g1.add_policy(policy, q2, 'export')

        # this will be blocked (exactly matches ^65100$)
        e1.add_route('192.168.100.0/24', aspath=[65100])
        # this will pass (does not match ^65100$)
        e1.add_route('192.168.200.0/24', aspath=[65001, 65100, 65010])

        for c in [e1, q1, q2]:
            g1.wait_for(BGP_FSM_ESTABLISHED, c)

    @staticmethod
    def check(env):
        # same check function as previous No.1 scenario
        lookup_scenario("ImportPolicy").check(env)

    @staticmethod
    def executor(env):
        lookup_scenario("ImportPolicyAsPathOnlyCondition").boot(env)
        lookup_scenario("ImportPolicyAsPathOnlyCondition").setup(env)
        lookup_scenario("ImportPolicyAsPathOnlyCondition").check(env)


@register_scenario
class ImportPolicyAsPathMismatchCondition(object):
    """
    No.14 aspath condition mismatch import-policy test
                                   -------------------------------
    exabgp ->(aspath=[...,65090])->| -> q1-rib -> q1-adj-rib-out | --> q1
                                   |                             |
                                   | -> q2-rib -> q2-adj-rib-out | --> q2
                                   -------------------------------
    This case check if policy passes the path to e1 because of condition mismatch.
    """
    @staticmethod
    def boot(env):
        lookup_scenario("ImportPolicy").boot(env)

    @staticmethod
    def setup(env):
        g1 = env.g1
        e1 = env.e1
        q1 = env.q1
        q2 = env.q2
        cs0 = {'community-sets': [{'community-set-name': 'cs0',
                                   'community-list': ['65100:10']}]}

        g1.set_bgp_defined_set(cs0)

        st0 = {'name': 'st0',
               'conditions': {'bgp-conditions': {'match-community-set': {'community-set': 'cs0'}}},
               'actions': {'route-disposition': 'reject-route'}}

        policy = {'name': 'policy0',
                  'statements': [st0]}
        g1.add_policy(policy, q2, 'export')

        # this will be blocked (no community match, but condition mismatch means it passes)
        e1.add_route('192.168.100.0/24', aspath=[65100, 65090])
        # this will pass (no community match)
        e1.add_route('192.168.200.0/24', aspath=[65001, 65100, 65010])

        for c in [e1, q1, q2]:
            g1.wait_for(BGP_FSM_ESTABLISHED, c)

    @staticmethod
    def check(env):
        g1 = env.g1
        # e1 = env.e1
        q1 = env.q1
        q2 = env.q2
        wait_for(lambda: len(g1.get_local_rib(q1)) == 2)
        wait_for(lambda: len(g1.get_adj_rib_out(q1)) == 2)
        wait_for(lambda: len(q1.get_global_rib()) == 2)
        wait_for(lambda: len(g1.get_local_rib(q2)) == 2)
        wait_for(lambda: len(g1.get_adj_rib_out(q2)) == 2)
        wait_for(lambda: len(q2.get_global_rib()) == 2)

    @staticmethod
    def executor(env):
        lookup_scenario("ImportPolicyAsPathMismatchCondition").boot(env)
        lookup_scenario("ImportPolicyAsPathMismatchCondition").setup(env)
        lookup_scenario("ImportPolicyAsPathMismatchCondition").check(env)


@register_scenario
class ImportPolicyCommunityCondition(object):
    """
    No.15 community condition import-policy test
                                --------------------------------
    e1 ->(community=65100:10)-> | ->  q1-rib -> q1-adj-rib-out | --> q1
                                |                              |
                                | ->x q2-rib                   |
                                --------------------------------
    """
    @staticmethod
    def boot(env):
        lookup_scenario("ImportPolicy").boot(env)

    @staticmethod
    def setup(env):
        g1 = env.g1
        e1 = env.e1
        q1 = env.q1
        q2 = env.q2
        cs0 = {'community-sets': [{'community-set-name': 'cs0', 'community-list': ['65100:10']}]}

        g1.set_bgp_defined_set(cs0)

        st0 = {'name': 'st0',
               'conditions': {'bgp-conditions': {'match-community-set': {'community-set': 'cs0'}}},
               'actions': {'route-disposition': 'reject-route'}}

        policy = {'name': 'policy0',
                  'statements': [st0]}
        g1.add_policy(policy, q2, 'export')

        # this will be blocked
        e1.add_route('192.168.100.0/24', community=['65100:10'])
        # this will pass
        e1.add_route('192.168.200.0/24', community=['65100:20'])

        for c in [e1, q1, q2]:
            g1.wait_for(BGP_FSM_ESTABLISHED, c)

    @staticmethod
    def check(env):
        lookup_scenario("ImportPolicy").check(env)

    @staticmethod
    def executor(env):
        lookup_scenario("ImportPolicyCommunityCondition").boot(env)
        lookup_scenario("ImportPolicyCommunityCondition").setup(env)
        lookup_scenario("ImportPolicyCommunityCondition").check(env)


@register_scenario
class ImportPolicyCommunityRegexp(object):
    """
    No.16 community condition regexp import-policy test
                                --------------------------------
    e1 ->(community=65100:10)-> | ->  q1-rib -> q1-adj-rib-out | --> q1
                                |                              |
                                | ->x q2-rib                   |
                                --------------------------------
    """
    @staticmethod
    def boot(env):
        lookup_scenario("ImportPolicy").boot(env)

    @staticmethod
    def setup(env):
        g1 = env.g1
        e1 = env.e1
        q1 = env.q1
        q2 = env.q2
        cs0 = {'community-sets': [{'community-set-name': 'cs0', 'community-list': ['6[0-9]+:[0-9]+']}]}

        g1.set_bgp_defined_set(cs0)

        st0 = {'name': 'st0',
               'conditions': {'bgp-conditions': {'match-community-set': {'community-set': 'cs0'}}},
               'actions': {'route-disposition': 'reject-route'}}

        policy = {'name': 'policy0',
                  'statements': [st0]}
        g1.add_policy(policy, q2, 'export')

        # this will be blocked
        e1.add_route('192.168.100.0/24', community=['65100:10'])
        # this will pass
        e1.add_route('192.168.200.0/24', community=['55100:20'])

        for c in [e1, q1, q2]:
            g1.wait_for(BGP_FSM_ESTABLISHED, c)

    @staticmethod
    def check(env):
        lookup_scenario("ImportPolicy").check(env)

    @staticmethod
    def executor(env):
        lookup_scenario("ImportPolicyCommunityRegexp").boot(env)
        lookup_scenario("ImportPolicyCommunityRegexp").setup(env)
        lookup_scenario("ImportPolicyCommunityRegexp").check(env)


def community_exists(path, com):
    a, b = com.split(':')
    com = (int(a) << 16) + int(b)
    for a in path['attrs']:
        if a['type'] == BGP_ATTR_TYPE_COMMUNITIES and com in a['communities']:
            return True
    return False


@register_scenario
class ImportPolicyCommunityAction(object):
    """
    No.17 community add action import-policy test
                                -------------------------------
    e1 ->(community=65100:10)-> | -> q1-rib -> q1-adj-rib-out | ->(community=65100:10)->          q1
                                |                             |
                                | -> q2-rib -> q2-adj-rib-out | ->(community=65100:10,65100:20)-> q2
                                |    apply action             |
                                -------------------------------
    """
    @staticmethod
    def boot(env):
        lookup_scenario("ImportPolicy").boot(env)

    @staticmethod
    def setup(env):
        g1 = env.g1
        e1 = env.e1
        q1 = env.q1
        q2 = env.q2
        cs0 = {'community-sets': [{'community-set-name': 'cs0', 'community-list': ['65100:10']}]}

        g1.set_bgp_defined_set(cs0)

        st0 = {'name': 'st0',
               'conditions': {'bgp-conditions': {'match-community-set': {'community-set': 'cs0', 'match-set-options': 'any'}}},
               'actions': {'route-disposition': 'accept-route',
                           'bgp-actions': {
                               'set-community': {
                                   'options': 'add',
                                   'set-community-method': {'communities-list': ['65100:20']}}}}}

        policy = {'name': 'policy0',
                  'statements': [st0]}
        g1.add_policy(policy, q2, 'export')

        e1.add_route('192.168.100.0/24', community=['65100:10'])

        for c in [e1, q1, q2]:
            g1.wait_for(BGP_FSM_ESTABLISHED, c)

    @staticmethod
    def check(env):
        g1 = env.g1
        # e1 = env.e1
        q1 = env.q1
        q2 = env.q2
        wait_for(lambda: len(g1.get_local_rib(q1)) == 1)
        wait_for(lambda: len(g1.get_adj_rib_out(q1)) == 1)
        wait_for(lambda: len(q1.get_global_rib()) == 1)
        wait_for(lambda: len(g1.get_local_rib(q2)) == 1)
        wait_for(lambda: len(g1.get_adj_rib_out(q2)) == 1)
        wait_for(lambda: len(q2.get_global_rib()) == 1)

    @staticmethod
    def check2(env):
        g1 = env.g1
        q1 = env.q1
        q2 = env.q2
        path = g1.get_adj_rib_out(q1)[0]
        assert community_exists(path, '65100:10')
        assert not community_exists(path, '65100:20')
        path = g1.get_adj_rib_out(q2)[0]
        assert community_exists(path, '65100:10')
        assert community_exists(path, '65100:20')

    @staticmethod
    def executor(env):
        lookup_scenario("ImportPolicyCommunityAction").boot(env)
        lookup_scenario("ImportPolicyCommunityAction").setup(env)
        lookup_scenario("ImportPolicyCommunityAction").check(env)
        lookup_scenario("ImportPolicyCommunityAction").check2(env)


@register_scenario
class ImportPolicyCommunityReplace(object):
    """
    No.18 community replace action import-policy test
                                -------------------------------
    e1 ->(community=65100:10)-> | -> q1-rib -> q1-adj-rib-out | ->(community=65100:10)-> q1
                                |                             |
                                | -> q2-rib -> q2-adj-rib-out | ->(community=65100:20)-> q2
                                |    apply action             |
                                -------------------------------
    """
    @staticmethod
    def boot(env):
        lookup_scenario("ImportPolicy").boot(env)

    @staticmethod
    def setup(env):
        g1 = env.g1
        e1 = env.e1
        q1 = env.q1
        q2 = env.q2
        cs0 = {'community-sets': [{'community-set-name': 'cs0', 'community-list': ['65100:10']}]}

        g1.set_bgp_defined_set(cs0)

        st0 = {'name': 'st0',
               'conditions': {'bgp-conditions': {'match-community-set': {'community-set': 'cs0'}}},
               'actions': {'route-disposition': 'accept-route',
                           'bgp-actions': {
                               'set-community': {
                                   'options': 'REPLACE',
                                   'set-community-method': {'communities-list': ['65100:20']}}}}}

        policy = {'name': 'policy0',
                  'statements': [st0]}
        g1.add_policy(policy, q2, 'export')

        e1.add_route('192.168.100.0/24', community=['65100:10'])

        for c in [e1, q1, q2]:
            g1.wait_for(BGP_FSM_ESTABLISHED, c)

    @staticmethod
    def check(env):
        lookup_scenario('ImportPolicyCommunityAction').check(env)

    @staticmethod
    def check2(env):
        g1 = env.g1
        # e1 = env.e1
        q1 = env.q1
        q2 = env.q2
        path = g1.get_adj_rib_out(q1)[0]
        assert community_exists(path, '65100:10')
        assert not community_exists(path, '65100:20')
        path = g1.get_adj_rib_out(q2)[0]
        assert not community_exists(path, '65100:10')
        assert community_exists(path, '65100:20')

    @staticmethod
    def executor(env):
        lookup_scenario("ImportPolicyCommunityReplace").boot(env)
        lookup_scenario("ImportPolicyCommunityReplace").setup(env)
        lookup_scenario("ImportPolicyCommunityReplace").check(env)
        lookup_scenario("ImportPolicyCommunityReplace").check2(env)


@register_scenario
class ImportPolicyCommunityRemove(object):
    """
    No.19 community remove action import-policy test
                                -------------------------------
    e1 ->(community=65100:10)-> | -> q1-rib -> q1-adj-rib-out | ->(community=65100:10)-> q1
                                |                             |
                                | -> q2-rib -> q2-adj-rib-out | ->(community=null)->     q2
                                |    apply action             |
                                -------------------------------
    """
    @staticmethod
    def boot(env):
        lookup_scenario("ImportPolicy").boot(env)

    @staticmethod
    def setup(env):
        g1 = env.g1
        e1 = env.e1
        q1 = env.q1
        q2 = env.q2
        cs0 = {'community-sets': [{'community-set-name': 'cs0', 'community-list': ['65100:10']}]}

        g1.set_bgp_defined_set(cs0)

        st0 = {'name': 'st0',
               'conditions': {'bgp-conditions': {'match-community-set': {'community-set': 'cs0'}}},
               'actions': {'route-disposition': 'accept-route',
                           'bgp-actions': {
                               'set-community': {
                                   'options': 'REMOVE',
                                   'set-community-method': {'communities-list': ['65100:10', '65100:20']}}}}}

        policy = {'name': 'policy0',
                  'statements': [st0]}
        g1.add_policy(policy, q2, 'export')

        e1.add_route('192.168.100.0/24', community=['65100:10'])
        e1.add_route('192.168.110.0/24', community=['65100:10', '65100:20'])
        e1.add_route('192.168.120.0/24', community=['65100:10', '65100:30'])

        for c in [e1, q1, q2]:
            g1.wait_for(BGP_FSM_ESTABLISHED, c)

    @staticmethod
    def check(env):
        g1 = env.g1
        # e1 = env.e1
        q1 = env.q1
        q2 = env.q2
        wait_for(lambda: len(g1.get_local_rib(q1)) == 3)
        wait_for(lambda: len(g1.get_adj_rib_out(q1)) == 3)
        wait_for(lambda: len(q1.get_global_rib()) == 3)
        wait_for(lambda: len(g1.get_local_rib(q2)) == 3)
        wait_for(lambda: len(g1.get_adj_rib_out(q2)) == 3)
        wait_for(lambda: len(q2.get_global_rib()) == 3)

    @staticmethod
    def check2(env):
        g1 = env.g1
        # e1 = env.e1
        q1 = env.q1
        q2 = env.q2
        adj_out = g1.get_adj_rib_out(q1)
        for path in adj_out:
            assert community_exists(path, '65100:10')
            if path['nlri']['prefix'] == '192.168.110.0/24':
                assert community_exists(path, '65100:20')
            if path['nlri']['prefix'] == '192.168.120.0/24':
                assert community_exists(path, '65100:30')
        adj_out = g1.get_adj_rib_out(q2)
        for path in adj_out:
            assert not community_exists(path, '65100:10')
            if path['nlri']['prefix'] == '192.168.110.0/24':
                assert not community_exists(path, '65100:20')
            if path['nlri']['prefix'] == '192.168.120.0/24':
                assert community_exists(path, '65100:30')

    @staticmethod
    def executor(env):
        lookup_scenario("ImportPolicyCommunityRemove").boot(env)
        lookup_scenario("ImportPolicyCommunityRemove").setup(env)
        lookup_scenario("ImportPolicyCommunityRemove").check(env)
        lookup_scenario("ImportPolicyCommunityRemove").check2(env)


@register_scenario
class ImportPolicyCommunityNull(object):
    """
    No.20 community null action import-policy test
                                -------------------------------
    e1 ->(community=65100:10)-> | -> q1-rib -> q1-adj-rib-out | ->(community=65100:10)-> q1
                                |                             |
                                | -> q2-rib -> q2-adj-rib-out | ->(community=null)->     q2
                                |    apply action             |
                                -------------------------------
    """
    @staticmethod
    def boot(env):
        lookup_scenario("ImportPolicy").boot(env)

    @staticmethod
    def setup(env):
        g1 = env.g1
        e1 = env.e1
        q1 = env.q1
        q2 = env.q2
        cs0 = {'community-sets': [{'community-set-name': 'cs0', 'community-list': ['65100:10']}]}

        g1.set_bgp_defined_set(cs0)

        st0 = {'name': 'st0',
               'conditions': {'bgp-conditions': {'match-community-set': {'community-set': 'cs0'}}},
               'actions': {'route-disposition': 'accept-route',
                           'bgp-actions': {
                               'set-community': {
                                   'options': 'REPLACE',
                                   'set-community-method': {'communities-list': []}}}}}

        policy = {'name': 'policy0',
                  'statements': [st0]}
        g1.add_policy(policy, q2, 'export')

        e1.add_route('192.168.100.0/24', community=['65100:10'])
        e1.add_route('192.168.110.0/24', community=['65100:10', '65100:20'])
        e1.add_route('192.168.120.0/24', community=['65100:10', '65100:30'])

        for c in [e1, q1, q2]:
            g1.wait_for(BGP_FSM_ESTABLISHED, c)

    @staticmethod
    def check(env):
        lookup_scenario('ImportPolicyCommunityRemove').check(env)

    @staticmethod
    def check2(env):
        g1 = env.g1
        q1 = env.q1
        q2 = env.q2
        adj_out = g1.get_adj_rib_out(q1)
        for path in adj_out:
            assert community_exists(path, '65100:10')
            if path['nlri']['prefix'] == '192.168.110.0/24':
                assert community_exists(path, '65100:20')
            if path['nlri']['prefix'] == '192.168.120.0/24':
                assert community_exists(path, '65100:30')
        adj_out = g1.get_adj_rib_out(q2)
        for path in adj_out:
            assert not community_exists(path, '65100:10')
            if path['nlri']['prefix'] == '192.168.110.0/24':
                assert not community_exists(path, '65100:20')
            if path['nlri']['prefix'] == '192.168.120.0/24':
                assert not community_exists(path, '65100:30')

    @staticmethod
    def executor(env):
        lookup_scenario("ImportPolicyCommunityNull").boot(env)
        lookup_scenario("ImportPolicyCommunityNull").setup(env)
        lookup_scenario("ImportPolicyCommunityNull").check(env)
        lookup_scenario("ImportPolicyCommunityNull").check2(env)


@register_scenario
class ExportPolicyCommunityAdd(object):
    """
    No.21 community add action export-policy test
                                -------------------------------
    e1 ->(community=65100:10)-> | -> q1-rib -> q1-adj-rib-out | ->(community=65100:10)-> q1
                                |                             |
                                | -> q2-rib -> q2-adj-rib-out | ->(community=null)->     q2
                                |              apply action   |
                                -------------------------------
    """
    @staticmethod
    def boot(env):
        lookup_scenario('ImportPolicy').boot(env)

    @staticmethod
    def setup(env):
        g1 = env.g1
        e1 = env.e1
        q1 = env.q1
        q2 = env.q2
        cs0 = {'community-sets': [{'community-set-name': 'cs0', 'community-list': ['65100:10']}]}

        g1.set_bgp_defined_set(cs0)

        st0 = {'name': 'st0',
               'conditions': {'bgp-conditions': {'match-community-set': {'community-set': 'cs0'}}},
               'actions': {'route-disposition': 'accept-route',
                           'bgp-actions': {
                               'set-community': {
                                   'options': 'add',
                                   'set-community-method': {'communities-list': ['65100:20']}}}}}

        policy = {'name': 'policy0',
                  'statements': [st0]}
        g1.add_policy(policy, q2, 'export')

        e1.add_route('192.168.100.0/24', community=['65100:10'])

        for c in [e1, q1, q2]:
            g1.wait_for(BGP_FSM_ESTABLISHED, c)

    @staticmethod
    def check(env):
        lookup_scenario('ImportPolicyCommunityAction').check(env)

    @staticmethod
    def check2(env):
        g1 = env.g1
        q1 = env.q1
        q2 = env.q2

        adj_out = g1.get_adj_rib_out(q1)
        for path in adj_out:
            assert community_exists(path, '65100:10')
            assert not community_exists(path, '65100:20')

        local_rib = g1.get_local_rib(q2)
        for path in local_rib[0]['paths']:
            assert community_exists(path, '65100:10')
            assert not community_exists(path, '65100:20')

        adj_out = g1.get_adj_rib_out(q2)
        for path in adj_out:
            assert community_exists(path, '65100:10')
            assert community_exists(path, '65100:20')

    @staticmethod
    def executor(env):
        lookup_scenario("ExportPolicyCommunityAdd").boot(env)
        lookup_scenario("ExportPolicyCommunityAdd").setup(env)
        lookup_scenario("ExportPolicyCommunityAdd").check(env)
        lookup_scenario("ExportPolicyCommunityAdd").check2(env)


@register_scenario
class ExportPolicyCommunityReplace(object):
    """
    No.22 community replace action export-policy test
                                -------------------------------
    e1 ->(community=65100:10)-> | -> q1-rib -> q1-adj-rib-out | ->(community=65100:10)-> q1
                                |                             |
                                | -> q2-rib -> q2-adj-rib-out | ->(community=65100:20)-> q2
                                |              apply action   |
                                -------------------------------
    """
    @staticmethod
    def boot(env):
        lookup_scenario('ImportPolicy').boot(env)

    @staticmethod
    def setup(env):
        g1 = env.g1
        e1 = env.e1
        q1 = env.q1
        q2 = env.q2
        cs0 = {'community-sets': [{'community-set-name': 'cs0', 'community-list': ['65100:10']}]}

        g1.set_bgp_defined_set(cs0)

        st0 = {'name': 'st0',
               'conditions': {'bgp-conditions': {'match-community-set': {'community-set': 'cs0'}}},
               'actions': {'route-disposition': 'accept-route',
                           'bgp-actions': {
                               'set-community': {
                                   'options': 'REPLACE',
                                   'set-community-method': {'communities-list': ['65100:20']}}}}}

        policy = {'name': 'policy0',
                  'statements': [st0]}
        g1.add_policy(policy, q2, 'export')

        e1.add_route('192.168.100.0/24', community=['65100:10'])

        for c in [e1, q1, q2]:
            g1.wait_for(BGP_FSM_ESTABLISHED, c)

    @staticmethod
    def check(env):
        lookup_scenario('ImportPolicyCommunityAction').check(env)

    @staticmethod
    def check2(env):
        g1 = env.g1
        q1 = env.q1
        q2 = env.q2

        adj_out = g1.get_adj_rib_out(q1)
        for path in adj_out:
            assert community_exists(path, '65100:10')
            assert not community_exists(path, '65100:20')

        local_rib = g1.get_local_rib(q2)
        for path in local_rib[0]['paths']:
            assert community_exists(path, '65100:10')
            assert not community_exists(path, '65100:20')

        adj_out = g1.get_adj_rib_out(q2)
        for path in adj_out:
            assert not community_exists(path, '65100:10')
            assert community_exists(path, '65100:20')

    @staticmethod
    def executor(env):
        lookup_scenario("ExportPolicyCommunityReplace").boot(env)
        lookup_scenario("ExportPolicyCommunityReplace").setup(env)
        lookup_scenario("ExportPolicyCommunityReplace").check(env)
        lookup_scenario("ExportPolicyCommunityReplace").check2(env)


@register_scenario
class ExportPolicyCommunityRemove(object):
    """
    No.23 community replace action export-policy test
                                -------------------------------
    e1 ->(community=65100:10)-> | -> q1-rib -> q1-adj-rib-out | ->(community=65100:10)-> q1
                                |                             |
                                | -> q2-rib -> q2-adj-rib-out | ->(community=null)->     q2
                                |              apply action   |
                                -------------------------------
    """
    @staticmethod
    def boot(env):
        lookup_scenario('ImportPolicy').boot(env)

    @staticmethod
    def setup(env):
        g1 = env.g1
        e1 = env.e1
        q1 = env.q1
        q2 = env.q2
        cs0 = {'community-sets': [{'community-set-name': 'cs0', 'community-list': ['65100:10']}]}

        g1.set_bgp_defined_set(cs0)

        st0 = {'name': 'st0',
               'conditions': {'bgp-conditions': {'match-community-set': {'community-set': 'cs0'}}},
               'actions': {'route-disposition': 'accept-route',
                           'bgp-actions': {
                               'set-community': {
                                   'options': 'REMOVE',
                                   'set-community-method': {'communities-list': ['65100:20', '65100:30']}}}}}

        policy = {'name': 'policy0',
                  'statements': [st0]}
        g1.add_policy(policy, q2, 'export')

        e1.add_route('192.168.100.0/24', community=['65100:10', '65100:20', '65100:30'])

        for c in [e1, q1, q2]:
            g1.wait_for(BGP_FSM_ESTABLISHED, c)

    @staticmethod
    def check(env):
        lookup_scenario('ImportPolicyCommunityAction').check(env)

    @staticmethod
    def check2(env):
        g1 = env.g1
        q1 = env.q1
        q2 = env.q2

        adj_out = g1.get_adj_rib_out(q1)
        for path in adj_out:
            assert community_exists(path, '65100:10')
            assert community_exists(path, '65100:20')
            assert community_exists(path, '65100:30')

        local_rib = g1.get_local_rib(q2)
        for path in local_rib[0]['paths']:
            assert community_exists(path, '65100:10')
            assert community_exists(path, '65100:20')
            assert community_exists(path, '65100:30')

        adj_out = g1.get_adj_rib_out(q2)
        for path in adj_out:
            assert community_exists(path, '65100:10')
            assert not community_exists(path, '65100:20')
            assert not community_exists(path, '65100:30')

    @staticmethod
    def executor(env):
        lookup_scenario("ExportPolicyCommunityRemove").boot(env)
        lookup_scenario("ExportPolicyCommunityRemove").setup(env)
        lookup_scenario("ExportPolicyCommunityRemove").check(env)
        lookup_scenario("ExportPolicyCommunityRemove").check2(env)


@register_scenario
class ExportPolicyCommunityNull(object):
    """
    No.24 community null action export-policy test
                                -------------------------------
    e1 ->(community=65100:10)-> | -> q1-rib -> q1-adj-rib-out | ->(community=65100:10)-> q1
                                |                             |
                                | -> q2-rib -> q2-adj-rib-out | ->(community=null)->     q2
                                |              apply action   |
                                -------------------------------
    """
    @staticmethod
    def boot(env):
        lookup_scenario('ImportPolicy').boot(env)

    @staticmethod
    def setup(env):
        g1 = env.g1
        e1 = env.e1
        q1 = env.q1
        q2 = env.q2
        cs0 = {'community-sets': [{'community-set-name': 'cs0', 'community-list': ['65100:10']}]}

        g1.set_bgp_defined_set(cs0)

        st0 = {'name': 'st0',
               'conditions': {'bgp-conditions': {'match-community-set': {'community-set': 'cs0'}}},
               'actions': {'route-disposition': 'accept-route',
                           'bgp-actions': {
                               'set-community': {
                                   'options': 'REPLACE',
                                   'set-community-method': {'communities-list': []}}}}}

        policy = {'name': 'policy0',
                  'statements': [st0]}
        g1.add_policy(policy, q2, 'export')

        e1.add_route('192.168.100.0/24', community=['65100:10', '65100:20', '65100:30'])

        for c in [e1, q1, q2]:
            g1.wait_for(BGP_FSM_ESTABLISHED, c)

    @staticmethod
    def check(env):
        lookup_scenario('ImportPolicyCommunityAction').check(env)

    @staticmethod
    def check2(env):
        g1 = env.g1
        q1 = env.q1
        q2 = env.q2

        adj_out = g1.get_adj_rib_out(q1)
        for path in adj_out:
            assert community_exists(path, '65100:10')
            assert community_exists(path, '65100:20')
            assert community_exists(path, '65100:30')

        local_rib = g1.get_local_rib(q2)
        for path in local_rib[0]['paths']:
            assert community_exists(path, '65100:10')
            assert community_exists(path, '65100:20')
            assert community_exists(path, '65100:30')

        adj_out = g1.get_adj_rib_out(q2)
        for path in adj_out:
            assert not community_exists(path, '65100:10')
            assert not community_exists(path, '65100:20')
            assert not community_exists(path, '65100:30')

    @staticmethod
    def executor(env):
        lookup_scenario("ExportPolicyCommunityNull").boot(env)
        lookup_scenario("ExportPolicyCommunityNull").setup(env)
        lookup_scenario("ExportPolicyCommunityNull").check(env)
        lookup_scenario("ExportPolicyCommunityNull").check2(env)


def metric(path):
    for a in path['attrs']:
        if 'metric' in a:
            return a['metric']
    return -1


@register_scenario
class ImportPolicyMedReplace(object):
    """
    No.25 med replace action import-policy test
                     -------------------------------
    e1 ->(med=300)-> | -> q1-rib -> q1-adj-rib-out | ->(med=300)-> q1
                     |                             |
                     | -> q2-rib -> q2-adj-rib-out | ->(med=100)-> q2
                     |    apply action             |
                     -------------------------------
    """
    @staticmethod
    def boot(env):
        lookup_scenario('ImportPolicy').boot(env)

    @staticmethod
    def setup(env):
        g1 = env.g1
        e1 = env.e1
        q1 = env.q1
        q2 = env.q2
        st0 = {'name': 'st0',
               'actions': {'route-disposition': 'accept-route',
                           'bgp-actions': {'set-med': '100'}}}

        policy = {'name': 'policy0',
                  'statements': [st0]}
        g1.add_policy(policy, q2, 'export')

        e1.add_route('192.168.100.0/24', med=300)

        for c in [e1, q1, q2]:
            g1.wait_for(BGP_FSM_ESTABLISHED, c)

    @staticmethod
    def check(env):
        lookup_scenario('ImportPolicyCommunityAction').check(env)

    @staticmethod
    def check2(env):
        g1 = env.g1
        q1 = env.q1
        q2 = env.q2

        adj_out = g1.get_adj_rib_out(q1)
        assert metric(adj_out[0]) == 300

        local_rib = g1.get_local_rib(q2)
        assert metric(local_rib[0]['paths'][0]) == 300

        adj_out = g1.get_adj_rib_out(q2)
        assert metric(adj_out[0]) == 100

    @staticmethod
    def executor(env):
        lookup_scenario("ImportPolicyMedReplace").boot(env)
        lookup_scenario("ImportPolicyMedReplace").setup(env)
        lookup_scenario("ImportPolicyMedReplace").check(env)
        lookup_scenario("ImportPolicyMedReplace").check2(env)


@register_scenario
class ImportPolicyMedAdd(object):
    """
    No.26 med add action import-policy test
                     -------------------------------
    e1 ->(med=300)-> | -> q1-rib -> q1-adj-rib-out | ->(med=300)->     q1
                     |                             |
                     | -> q2-rib -> q2-adj-rib-out | ->(med=300+100)-> q2
                     |    apply action             |
                     -------------------------------
    """
    @staticmethod
    def boot(env):
        lookup_scenario('ImportPolicy').boot(env)

    @staticmethod
    def setup(env):
        g1 = env.g1
        e1 = env.e1
        q1 = env.q1
        q2 = env.q2
        st0 = {'name': 'st0',
               'actions': {'route-disposition': 'accept-route',
                           'bgp-actions': {'set-med': '+100'}}}

        policy = {'name': 'policy0',
                  'statements': [st0]}
        g1.add_policy(policy, q2, 'export')

        e1.add_route('192.168.100.0/24', med=300)

        for c in [e1, q1, q2]:
            g1.wait_for(BGP_FSM_ESTABLISHED, c)

    @staticmethod
    def check(env):
        lookup_scenario('ImportPolicyCommunityAction').check(env)

    @staticmethod
    def check2(env):
        g1 = env.g1
        q1 = env.q1
        q2 = env.q2

        adj_out = g1.get_adj_rib_out(q1)
        assert metric(adj_out[0]) == 300

        local_rib = g1.get_local_rib(q2)
        assert metric(local_rib[0]['paths'][0]) == 300

        adj_out = g1.get_adj_rib_out(q2)
        assert metric(adj_out[0]) == 400

    @staticmethod
    def executor(env):
        lookup_scenario("ImportPolicyMedAdd").boot(env)
        lookup_scenario("ImportPolicyMedAdd").setup(env)
        lookup_scenario("ImportPolicyMedAdd").check(env)
        lookup_scenario("ImportPolicyMedAdd").check2(env)


@register_scenario
class ImportPolicyMedSub(object):
    """
    No.27 med subtract action import-policy test
                     -------------------------------
    e1 ->(med=300)-> | -> q1-rib -> q1-adj-rib-out | ->(med=300)->     q1
                     |                             |
                     | -> q2-rib -> q2-adj-rib-out | ->(med=300-100)-> q2
                     |    apply action             |
                     -------------------------------
    """
    @staticmethod
    def boot(env):
        lookup_scenario('ImportPolicy').boot(env)

    @staticmethod
    def setup(env):
        g1 = env.g1
        e1 = env.e1
        q1 = env.q1
        q2 = env.q2
        st0 = {'name': 'st0',
               'actions': {'route-disposition': 'accept-route',
                           'bgp-actions': {'set-med': '-100'}}}

        policy = {'name': 'policy0',
                  'statements': [st0]}
        g1.add_policy(policy, q2, 'export')

        e1.add_route('192.168.100.0/24', med=300)

        for c in [e1, q1, q2]:
            g1.wait_for(BGP_FSM_ESTABLISHED, c)

    @staticmethod
    def check(env):
        lookup_scenario('ImportPolicyCommunityAction').check(env)

    @staticmethod
    def check2(env):
        g1 = env.g1
        q1 = env.q1
        q2 = env.q2

        adj_out = g1.get_adj_rib_out(q1)
        assert metric(adj_out[0]) == 300

        local_rib = g1.get_local_rib(q2)
        assert metric(local_rib[0]['paths'][0]) == 300

        adj_out = g1.get_adj_rib_out(q2)
        assert metric(adj_out[0]) == 200

    @staticmethod
    def executor(env):
        lookup_scenario("ImportPolicyMedSub").boot(env)
        lookup_scenario("ImportPolicyMedSub").setup(env)
        lookup_scenario("ImportPolicyMedSub").check(env)
        lookup_scenario("ImportPolicyMedSub").check2(env)


@register_scenario
class ExportPolicyMedReplace(object):
    """
    No.28 med replace action export-policy test
                     -------------------------------
    e1 ->(med=300)-> | -> q1-rib -> q1-adj-rib-out | ->(med=300)-> q1
                     |                             |
                     | -> q2-rib -> q2-adj-rib-out | ->(med=100)-> q2
                     |              apply action   |
                     -------------------------------
    """
    @staticmethod
    def boot(env):
        lookup_scenario('ImportPolicy').boot(env)

    @staticmethod
    def setup(env):
        g1 = env.g1
        e1 = env.e1
        q1 = env.q1
        q2 = env.q2
        st0 = {'name': 'st0',
               'actions': {'route-disposition': 'accept-route',
                           'bgp-actions': {'set-med': '100'}}}

        policy = {'name': 'policy0',
                  'statements': [st0]}
        g1.add_policy(policy, q2, 'export')

        e1.add_route('192.168.100.0/24', med=300)

        for c in [e1, q1, q2]:
            g1.wait_for(BGP_FSM_ESTABLISHED, c)

    @staticmethod
    def check(env):
        lookup_scenario('ImportPolicyCommunityAction').check(env)

    @staticmethod
    def check2(env):
        g1 = env.g1
        q1 = env.q1
        q2 = env.q2

        adj_out = g1.get_adj_rib_out(q1)
        assert metric(adj_out[0]) == 300

        local_rib = g1.get_local_rib(q2)
        assert metric(local_rib[0]['paths'][0]) == 300

        adj_out = g1.get_adj_rib_out(q2)
        assert metric(adj_out[0]) == 100

    @staticmethod
    def executor(env):
        lookup_scenario("ExportPolicyMedReplace").boot(env)
        lookup_scenario("ExportPolicyMedReplace").setup(env)
        lookup_scenario("ExportPolicyMedReplace").check(env)
        lookup_scenario("ExportPolicyMedReplace").check2(env)


@register_scenario
class ExportPolicyMedAdd(object):
    """
    No.29 med add action export-policy test
                     -------------------------------
    e1 ->(med=300)-> | -> q1-rib -> q1-adj-rib-out | ->(med=300)->     q1
                     |                             |
                     | -> q2-rib -> q2-adj-rib-out | ->(med=300+100)-> q2
                     |              apply action   |
                     -------------------------------
    """
    @staticmethod
    def boot(env):
        lookup_scenario('ImportPolicy').boot(env)

    @staticmethod
    def setup(env):
        g1 = env.g1
        e1 = env.e1
        q1 = env.q1
        q2 = env.q2
        st0 = {'name': 'st0',
               'actions': {'route-disposition': 'accept-route',
                           'bgp-actions': {'set-med': '+100'}}}

        policy = {'name': 'policy0',
                  'statements': [st0]}
        g1.add_policy(policy, q2, 'export')

        e1.add_route('192.168.100.0/24', med=300)

        for c in [e1, q1, q2]:
            g1.wait_for(BGP_FSM_ESTABLISHED, c)

    @staticmethod
    def check(env):
        lookup_scenario('ImportPolicyCommunityAction').check(env)

    @staticmethod
    def check2(env):
        g1 = env.g1
        q1 = env.q1
        q2 = env.q2

        adj_out = g1.get_adj_rib_out(q1)
        assert metric(adj_out[0]) == 300

        local_rib = g1.get_local_rib(q2)
        assert metric(local_rib[0]['paths'][0]) == 300

        adj_out = g1.get_adj_rib_out(q2)
        assert metric(adj_out[0]) == 400

    @staticmethod
    def executor(env):
        lookup_scenario("ExportPolicyMedAdd").boot(env)
        lookup_scenario("ExportPolicyMedAdd").setup(env)
        lookup_scenario("ExportPolicyMedAdd").check(env)
        lookup_scenario("ExportPolicyMedAdd").check2(env)


@register_scenario
class ExportPolicyMedSub(object):
    """
    No.30 med subtract action export-policy test
                     -------------------------------
    e1 ->(med=300)-> | -> q1-rib -> q1-adj-rib-out | ->(med=300)->     q1
                     |                             |
                     | -> q2-rib -> q2-adj-rib-out | ->(med=300-100)-> q2
                     |              apply action   |
                     -------------------------------
    """
    @staticmethod
    def boot(env):
        lookup_scenario('ImportPolicy').boot(env)

    @staticmethod
    def setup(env):
        g1 = env.g1
        e1 = env.e1
        q1 = env.q1
        q2 = env.q2
        st0 = {'name': 'st0',
               'actions': {'route-disposition': 'accept-route',
                           'bgp-actions': {'set-med': '-100'}}}

        policy = {'name': 'policy0',
                  'statements': [st0]}
        g1.add_policy(policy, q2, 'export')

        e1.add_route('192.168.100.0/24', med=300)

        for c in [e1, q1, q2]:
            g1.wait_for(BGP_FSM_ESTABLISHED, c)

    @staticmethod
    def check(env):
        lookup_scenario('ImportPolicyCommunityAction').check(env)

    @staticmethod
    def check2(env):
        g1 = env.g1
        q1 = env.q1
        q2 = env.q2

        adj_out = g1.get_adj_rib_out(q1)
        assert metric(adj_out[0]) == 300

        local_rib = g1.get_local_rib(q2)
        assert metric(local_rib[0]['paths'][0]) == 300

        adj_out = g1.get_adj_rib_out(q2)
        assert metric(adj_out[0]) == 200

    @staticmethod
    def executor(env):
        lookup_scenario("ExportPolicyMedSub").boot(env)
        lookup_scenario("ExportPolicyMedSub").setup(env)
        lookup_scenario("ExportPolicyMedSub").check(env)
        lookup_scenario("ExportPolicyMedSub").check2(env)


@register_scenario
class ExportPolicyAsPathPrepend(object):
    """
    No.37 aspath prepend action export
                            --------------------------------
    e1 ->(aspath=[65001])-> | ->  p1-rib -> p1-adj-rib-out | -> p1
                            |                              |
                            | ->  p2-rib -> p2-adj-rib-out | -> p2
                            |               apply action   |
                            --------------------------------
    """
    @staticmethod
    def boot(env):
        lookup_scenario("ImportPolicy").boot(env)

    @staticmethod
    def setup(env):
        g1 = env.g1
        e1 = env.e1
        q1 = env.q1
        q2 = env.q2

        p0 = {'ip-prefix': '192.168.20.0/24'}

        ps0 = {'prefix-set-name': 'ps0',
               'prefix-list': [p0]}
        g1.set_prefix_set(ps0)

        st0 = {'name': 'st0',
               'conditions': {'match-prefix-set': {'prefix-set': ps0['prefix-set-name']}},
               'actions': {'route-disposition': 'accept-route',
                           'bgp-actions': {'set-as-path-prepend': {'repeat-n': 5, 'as': "65005"}}}}

        policy = {'name': 'policy0',
                  'statements': [st0]}
        g1.add_policy(policy, q2, 'export')

        e1.add_route('192.168.20.0/24')
        e1.add_route('192.168.200.0/24')

        for c in [e1, q1, q2]:
            g1.wait_for(BGP_FSM_ESTABLISHED, c)

    @staticmethod
    def check(env):
        g1 = env.g1
        e1 = env.e1
        q1 = env.q1
        q2 = env.q2
        wait_for(lambda: len(g1.get_adj_rib_in(e1)) == 2)
        wait_for(lambda: len(g1.get_local_rib(q1)) == 2)
        wait_for(lambda: len(g1.get_adj_rib_out(q1)) == 2)
        wait_for(lambda: len(q1.get_global_rib()) == 2)
        wait_for(lambda: len(g1.get_local_rib(q2)) == 2)
        wait_for(lambda: len(g1.get_adj_rib_out(q2)) == 2)
        wait_for(lambda: len(q2.get_global_rib()) == 2)

    @staticmethod
    def check2(env):
        g1 = env.g1
        e1 = env.e1
        q1 = env.q1
        q2 = env.q2

        path = g1.get_adj_rib_out(q1, prefix='192.168.20.0/24')[0]
        assert path['aspath'] == [e1.asn]

        path = g1.get_adj_rib_out(q1, prefix='192.168.200.0/24')[0]
        assert path['aspath'] == [e1.asn]

        path = g1.get_local_rib(q2, prefix='192.168.20.0/24')[0]['paths'][0]
        assert path['aspath'] == [e1.asn]

        path = g1.get_adj_rib_out(q2, prefix='192.168.20.0/24')[0]
        assert path['aspath'] == ([65005] * 5 + [e1.asn])

        path = g1.get_adj_rib_out(q2, prefix='192.168.200.0/24')[0]
        assert path['aspath'] == [e1.asn]

    @staticmethod
    def executor(env):
        lookup_scenario("ExportPolicyAsPathPrepend").boot(env)
        lookup_scenario("ExportPolicyAsPathPrepend").setup(env)
        lookup_scenario("ExportPolicyAsPathPrepend").check(env)
        lookup_scenario("ExportPolicyAsPathPrepend").check2(env)


@register_scenario
class ImportPolicyAsPathPrependLastAS(object):
    """
    No.38 aspath prepend action lastas import
                            --------------------------------
    e1 ->(aspath=[65001])-> | ->  p1-rib -> p1-adj-rib-out | -> p1
                            |                              |
                            | ->  p2-rib -> p2-adj-rib-out | -> p2
                            |     apply action             |
                            --------------------------------
    """
    @staticmethod
    def boot(env):
        lookup_scenario("ImportPolicy").boot(env)

    @staticmethod
    def setup(env):
        g1 = env.g1
        e1 = env.e1
        q1 = env.q1
        q2 = env.q2

        p0 = {'ip-prefix': '192.168.20.0/24'}

        ps0 = {'prefix-set-name': 'ps0',
               'prefix-list': [p0]}
        g1.set_prefix_set(ps0)

        st0 = {'name': 'st0',
               'conditions': {'match-prefix-set': {'prefix-set': ps0['prefix-set-name']}},
               'actions': {'route-disposition': 'accept-route',
                           'bgp-actions': {'set-as-path-prepend': {'repeat-n': 5, 'as': "last-as"}}}}

        policy = {'name': 'policy0',
                  'statements': [st0]}
        g1.add_policy(policy, q2, 'export')

        e1.add_route('192.168.20.0/24')
        e1.add_route('192.168.200.0/24')

        for c in [e1, q1, q2]:
            g1.wait_for(BGP_FSM_ESTABLISHED, c)

    @staticmethod
    def check(env):
        lookup_scenario('ExportPolicyAsPathPrepend').check(env)

    @staticmethod
    def check2(env):
        g1 = env.g1
        e1 = env.e1
        q1 = env.q1
        q2 = env.q2

        path = g1.get_adj_rib_out(q1, prefix='192.168.20.0/24')[0]
        assert path['aspath'] == [e1.asn]

        path = g1.get_adj_rib_out(q1, prefix='192.168.200.0/24')[0]
        assert path['aspath'] == [e1.asn]

        path = g1.get_local_rib(q2, prefix='192.168.20.0/24')[0]['paths'][0]
        assert path['aspath'] == [e1.asn]

        path = g1.get_adj_rib_out(q2, prefix='192.168.20.0/24')[0]
        assert path['aspath'] == ([e1.asn] * 5 + [e1.asn])

        path = g1.get_adj_rib_out(q2, prefix='192.168.200.0/24')[0]
        assert path['aspath'] == [e1.asn]

    @staticmethod
    def executor(env):
        lookup_scenario("ImportPolicyAsPathPrependLastAS").boot(env)
        lookup_scenario("ImportPolicyAsPathPrependLastAS").setup(env)
        lookup_scenario("ImportPolicyAsPathPrependLastAS").check(env)
        lookup_scenario("ImportPolicyAsPathPrependLastAS").check2(env)


@register_scenario
class ExportPolicyAsPathPrependLastAS(object):
    """
    No.39 aspath prepend action lastas export
                            --------------------------------
    e1 ->(aspath=[65001])-> | ->  p1-rib -> p1-adj-rib-out | -> p1
                            |                              |
                            | ->  p2-rib -> p2-adj-rib-out | -> p2
                            |     apply action             |
                            --------------------------------
    """
    @staticmethod
    def boot(env):
        lookup_scenario("ImportPolicy").boot(env)

    @staticmethod
    def setup(env):
        g1 = env.g1
        e1 = env.e1
        q1 = env.q1
        q2 = env.q2

        p0 = {'ip-prefix': '192.168.20.0/24'}

        ps0 = {'prefix-set-name': 'ps0',
               'prefix-list': [p0]}
        g1.set_prefix_set(ps0)

        st0 = {'name': 'st0',
               'conditions': {'match-prefix-set': {'prefix-set': ps0['prefix-set-name']}},
               'actions': {'route-disposition': 'accept-route',
                           'bgp-actions': {'set-as-path-prepend': {'repeat-n': 5, 'as': "last-as"}}}}

        policy = {'name': 'policy0',
                  'statements': [st0]}
        g1.add_policy(policy, q2, 'export')

        e1.add_route('192.168.20.0/24')
        e1.add_route('192.168.200.0/24')

        for c in [e1, q1, q2]:
            g1.wait_for(BGP_FSM_ESTABLISHED, c)

    @staticmethod
    def check(env):
        lookup_scenario('ExportPolicyAsPathPrepend').check(env)

    @staticmethod
    def check2(env):
        g1 = env.g1
        e1 = env.e1
        q1 = env.q1
        q2 = env.q2

        path = g1.get_adj_rib_out(q1, prefix='192.168.20.0/24')[0]
        assert path['aspath'] == [e1.asn]

        path = g1.get_adj_rib_out(q1, prefix='192.168.200.0/24')[0]
        assert path['aspath'] == [e1.asn]

        path = g1.get_local_rib(q2, prefix='192.168.20.0/24')[0]['paths'][0]
        assert path['aspath'] == [e1.asn]

        path = g1.get_adj_rib_out(q2, prefix='192.168.20.0/24')[0]
        assert path['aspath'] == ([e1.asn] * 5 + [e1.asn])

        path = g1.get_adj_rib_out(q2, prefix='192.168.200.0/24')[0]
        assert path['aspath'] == [e1.asn]

    @staticmethod
    def executor(env):
        lookup_scenario("ExportPolicyAsPathPrependLastAS").boot(env)
        lookup_scenario("ExportPolicyAsPathPrependLastAS").setup(env)
        lookup_scenario("ExportPolicyAsPathPrependLastAS").check(env)
        lookup_scenario("ExportPolicyAsPathPrependLastAS").check2(env)


@register_scenario
class ImportPolicyExCommunityOriginCondition(object):
    """
    No.40 extended community origin condition import
                                                 --------------------------------
    e1 ->(extcommunity=origin:65001.65100:200)-> | ->  q1-rib -> q1-adj-rib-out | --> q1
                                                 |                              |
                                                 | ->x q2-rib                   |
                                                 --------------------------------
    """
    @staticmethod
    def boot(env):
        lookup_scenario("ImportPolicy").boot(env)

    @staticmethod
    def setup(env):
        g1 = env.g1
        e1 = env.e1
        q1 = env.q1
        q2 = env.q2

        es0 = {'ext-community-sets': [{'ext-community-set-name': 'es0',
                                       'ext-community-list': ['SoO:65001.65100:200']}]}

        g1.set_bgp_defined_set(es0)

        st0 = {'name': 'st0',
               'conditions': {'bgp-conditions': {'match-ext-community-set': {'ext-community-set': 'es0'}}},
               'actions': {'route-disposition': 'reject-route'}}

        policy = {'name': 'policy0',
                  'statements': [st0]}
        g1.add_policy(policy, q2, 'export')

        e1.add_route('192.168.20.0/24', extendedcommunity='origin:{0}:200'.format((65001 << 16) + 65100))
        e1.add_route('192.168.200.0/24', extendedcommunity='origin:{0}:100'.format((65001 << 16) + 65200))

        for c in [e1, q1, q2]:
            g1.wait_for(BGP_FSM_ESTABLISHED, c)

    @staticmethod
    def check(env):
        lookup_scenario("ImportPolicy").check(env)

    @staticmethod
    def executor(env):
        lookup_scenario("ImportPolicyExCommunityOriginCondition").boot(env)
        lookup_scenario("ImportPolicyExCommunityOriginCondition").setup(env)
        lookup_scenario("ImportPolicyExCommunityOriginCondition").check(env)


@register_scenario
class ImportPolicyExCommunityTargetCondition(object):
    """
    No.41 extended community origin condition import
                                           --------------------------------
    e1 ->(extcommunity=target:65010:320)-> | ->  q1-rib -> q1-adj-rib-out | --> q1
                                           |                              |
                                           | ->x q2-rib                   |
                                           --------------------------------
    """
    @staticmethod
    def boot(env):
        lookup_scenario('ImportPolicy').boot(env)

    @staticmethod
    def setup(env):
        g1 = env.g1
        e1 = env.e1
        q1 = env.q1
        q2 = env.q2

        es0 = {'ext-community-sets': [{'ext-community-set-name': 'es0',
                                       'ext-community-list': ['RT:6[0-9]+:3[0-9]+']}]}

        g1.set_bgp_defined_set(es0)

        st0 = {'name': 'st0',
               'conditions': {'bgp-conditions': {'match-ext-community-set': {'ext-community-set': 'es0'}}},
               'actions': {'route-disposition': 'reject-route'}}

        policy = {'name': 'policy0',
                  'statements': [st0]}
        g1.add_policy(policy, q2, 'export')

        e1.add_route('192.168.20.0/24', extendedcommunity='target:65010:320')
        e1.add_route('192.168.200.0/24', extendedcommunity='target:55000:320')

        for c in [e1, q1, q2]:
            g1.wait_for(BGP_FSM_ESTABLISHED, c)

    @staticmethod
    def check(env):
        lookup_scenario("ImportPolicy").check(env)

    @staticmethod
    def executor(env):
        lookup_scenario("ImportPolicyExCommunityTargetCondition").boot(env)
        lookup_scenario("ImportPolicyExCommunityTargetCondition").setup(env)
        lookup_scenario("ImportPolicyExCommunityTargetCondition").check(env)


def ext_community_exists(path, extcomm):
    typ = extcomm.split(':')[0]
    value = ':'.join(extcomm.split(':')[1:])
    for a in path['attrs']:
        if a['type'] == BGP_ATTR_TYPE_EXTENDED_COMMUNITIES:
            for c in a['value']:
                if typ == 'RT' and c['type'] == 0 and c['subtype'] == 2 and c['value'] == value:
                    return True
    return False


@register_scenario
class ImportPolicyExCommunityAdd(object):
    """
    No.43 extended community add action import-policy test
                               ---------------------------------
    e1 ->(extcommunity=none) ->| ->  q1-rib ->  q1-adj-rib-out | --> q1
                               |                               |
                               | ->  q2-rib ->  q2-adj-rib-out | --> q2
                               |     add ext-community         |
                               ---------------------------------
    """
    @staticmethod
    def boot(env):
        lookup_scenario('ImportPolicy').boot(env)

    @staticmethod
    def setup(env):
        g1 = env.g1
        e1 = env.e1
        q1 = env.q1
        q2 = env.q2

        p0 = {'ip-prefix': '192.168.10.0/24'}

        ps0 = {'prefix-set-name': 'ps0',
               'prefix-list': [p0]}
        g1.set_prefix_set(ps0)

        st0 = {
            'name': 'st0',
            'conditions': {
                'match-prefix-set': {
                    'prefix-set': ps0['prefix-set-name']
                }
            },
            'actions': {
                'route-disposition': 'accept-route',
                'bgp-actions': {
                    'set-ext-community': {
                        'options': 'add',
                        'set-ext-community-method': {
                            'communities-list': ['rt:65000:1'],
                        }
                    },
                }
            }
        }

        policy = {'name': 'policy0',
                  'statements': [st0]}
        g1.add_policy(policy, q2, 'export')

        e1.add_route('192.168.10.0/24')

        for c in [e1, q1, q2]:
            g1.wait_for(BGP_FSM_ESTABLISHED, c)

    @staticmethod
    def check(env):
        lookup_scenario('ImportPolicyCommunityAction').check(env)

    @staticmethod
    def check2(env):
        g1 = env.g1
        # e1 = env.e1
        q1 = env.q1
        q2 = env.q2
        path = g1.get_adj_rib_out(q1)[0]
        assert not ext_community_exists(path, 'RT:65000:1')
        path = g1.get_adj_rib_out(q2)[0]
        assert ext_community_exists(path, 'RT:65000:1')

    @staticmethod
    def executor(env):
        lookup_scenario("ImportPolicyExCommunityAdd").boot(env)
        lookup_scenario("ImportPolicyExCommunityAdd").setup(env)
        lookup_scenario("ImportPolicyExCommunityAdd").check(env)
        lookup_scenario("ImportPolicyExCommunityAdd").check2(env)


@register_scenario
class ImportPolicyExCommunityAdd2(object):
    """
    No.44 extended community add action import-policy test
                                      --------------------------------
    e1 ->(extcommunity=RT:65000:1) -> | ->  q1-rib -> q1-adj-rib-out | --> q1
                                      |                              |
                                      | ->  q2-rib -> q2-adj-rib-out | --> q2
                                      |     add ext-community        |
                                      --------------------------------
    """
    @staticmethod
    def boot(env):
        lookup_scenario('ImportPolicy').boot(env)

    @staticmethod
    def setup(env):
        g1 = env.g1
        e1 = env.e1
        q1 = env.q1
        q2 = env.q2

        p0 = {'ip-prefix': '192.168.10.0/24'}

        ps0 = {'prefix-set-name': 'ps0',
               'prefix-list': [p0]}
        g1.set_prefix_set(ps0)

        st0 = {
            'name': 'st0',
            'conditions': {
                'match-prefix-set': {
                    'prefix-set': ps0['prefix-set-name']
                }
            },
            'actions': {
                'route-disposition': 'accept-route',
                'bgp-actions': {
                    'set-ext-community': {
                        'options': 'add',
                        'set-ext-community-method': {
                            'communities-list': ['rt:65100:100'],
                        }
                    },
                }
            }
        }

        policy = {'name': 'policy0',
                  'statements': [st0]}
        g1.add_policy(policy, q2, 'export')

        e1.add_route('192.168.10.0/24', extendedcommunity='target:65000:1')

        for c in [e1, q1, q2]:
            g1.wait_for(BGP_FSM_ESTABLISHED, c)

    @staticmethod
    def check(env):
        lookup_scenario('ImportPolicyCommunityAction').check(env)

    @staticmethod
    def check2(env):
        g1 = env.g1
        # e1 = env.e1
        q1 = env.q1
        q2 = env.q2
        path = g1.get_adj_rib_out(q1)[0]
        assert ext_community_exists(path, 'RT:65000:1')
        assert not ext_community_exists(path, 'RT:65100:100')
        path = g1.get_local_rib(q2)[0]['paths'][0]
        assert ext_community_exists(path, 'RT:65000:1')
        assert not ext_community_exists(path, 'RT:65100:100')
        path = g1.get_adj_rib_out(q2)[0]
        assert ext_community_exists(path, 'RT:65000:1')
        assert ext_community_exists(path, 'RT:65100:100')

    @staticmethod
    def executor(env):
        lookup_scenario("ImportPolicyExCommunityAdd2").boot(env)
        lookup_scenario("ImportPolicyExCommunityAdd2").setup(env)
        lookup_scenario("ImportPolicyExCommunityAdd2").check(env)
        lookup_scenario("ImportPolicyExCommunityAdd2").check2(env)


@register_scenario
class ImportPolicyExCommunityMultipleAdd(object):
    """
    No.45 extended community add action multiple import-policy test
                                   ---------------------------------------
    exabgp ->(extcommunity=none) ->| ->  peer1-rib ->  peer1-adj-rib-out | --> peer1
                                   |                                     |
                                   | ->  peer2-rib ->  peer2-adj-rib-out | --> peer2
                                   |     add ext-community               |
                                   ---------------------------------------
    """
    @staticmethod
    def boot(env):
        lookup_scenario('ImportPolicy').boot(env)

    @staticmethod
    def setup(env):
        g1 = env.g1
        e1 = env.e1
        q1 = env.q1
        q2 = env.q2

        p0 = {'ip-prefix': '192.168.10.0/24'}

        ps0 = {'prefix-set-name': 'ps0',
               'prefix-list': [p0]}
        g1.set_prefix_set(ps0)

        st0 = {
            'name': 'st0',
            'conditions': {
                'match-prefix-set': {
                    'prefix-set': ps0['prefix-set-name']
                }
            },
            'actions': {
                'route-disposition': 'accept-route',
                'bgp-actions': {
                    'set-ext-community': {
                        'options': 'add',
                        'set-ext-community-method': {
                            'communities-list': ['rt:65100:100', 'rt:100:100'],
                        }
                    },
                }
            }
        }

        policy = {'name': 'policy0',
                  'statements': [st0]}
        g1.add_policy(policy, q2, 'export')

        e1.add_route('192.168.10.0/24')

        for c in [e1, q1, q2]:
            g1.wait_for(BGP_FSM_ESTABLISHED, c)

    @staticmethod
    def check(env):
        lookup_scenario('ImportPolicyCommunityAction').check(env)

    @staticmethod
    def check2(env):
        g1 = env.g1
        # e1 = env.e1
        q1 = env.q1
        q2 = env.q2
        path = g1.get_adj_rib_out(q1)[0]
        assert not ext_community_exists(path, 'RT:65100:100')
        assert not ext_community_exists(path, 'RT:100:100')
        path = g1.get_local_rib(q2)[0]['paths'][0]
        assert not ext_community_exists(path, 'RT:65100:100')
        assert not ext_community_exists(path, 'RT:100:100')
        path = g1.get_adj_rib_out(q2)[0]
        assert ext_community_exists(path, 'RT:65100:100')
        assert ext_community_exists(path, 'RT:100:100')

    @staticmethod
    def executor(env):
        lookup_scenario("ImportPolicyExCommunityMultipleAdd").boot(env)
        lookup_scenario("ImportPolicyExCommunityMultipleAdd").setup(env)
        lookup_scenario("ImportPolicyExCommunityMultipleAdd").check(env)
        lookup_scenario("ImportPolicyExCommunityMultipleAdd").check2(env)


@register_scenario
class ExportPolicyExCommunityAdd(object):
    """
    No.46 extended comunity add action export-policy test
                               ------------------------------------
    e1 ->(extcommunity=none) ->| ->  q1-rib ->  q1-adj-rib-out    | --> q1
                               |                                  |
                               | ->  q2-rib ->  q2-adj-rib-out    | --> q2
                               |                add ext-community |
                               ------------------------------------
    """
    @staticmethod
    def boot(env):
        lookup_scenario('ImportPolicy').boot(env)

    @staticmethod
    def setup(env):
        g1 = env.g1
        e1 = env.e1
        q1 = env.q1
        q2 = env.q2

        p0 = {'ip-prefix': '192.168.10.0/24'}

        ps0 = {'prefix-set-name': 'ps0',
               'prefix-list': [p0]}
        g1.set_prefix_set(ps0)

        st0 = {
            'name': 'st0',
            'conditions': {
                'match-prefix-set': {
                    'prefix-set': ps0['prefix-set-name']
                }
            },
            'actions': {
                'route-disposition': 'accept-route',
                'bgp-actions': {
                    'set-ext-community': {
                        'options': 'add',
                        'set-ext-community-method': {
                            'communities-list': ['rt:65000:1'],
                        }
                    },
                }
            }
        }

        policy = {'name': 'policy0',
                  'statements': [st0]}
        g1.add_policy(policy, q2, 'export')

        e1.add_route('192.168.10.0/24')

        for c in [e1, q1, q2]:
            g1.wait_for(BGP_FSM_ESTABLISHED, c)

    @staticmethod
    def check(env):
        lookup_scenario('ImportPolicyCommunityAction').check(env)

    @staticmethod
    def check2(env):
        g1 = env.g1
        # e1 = env.e1
        q1 = env.q1
        q2 = env.q2
        path = g1.get_adj_rib_out(q1)[0]
        assert not ext_community_exists(path, 'RT:65000:1')
        path = g1.get_local_rib(q2)[0]['paths'][0]
        assert not ext_community_exists(path, 'RT:65000:1')
        path = g1.get_adj_rib_out(q2)[0]
        assert ext_community_exists(path, 'RT:65000:1')

    @staticmethod
    def executor(env):
        lookup_scenario("ExportPolicyExCommunityAdd").boot(env)
        lookup_scenario("ExportPolicyExCommunityAdd").setup(env)
        lookup_scenario("ExportPolicyExCommunityAdd").check(env)
        lookup_scenario("ExportPolicyExCommunityAdd").check2(env)


@register_scenario
class ImportPolicyRejectImplicitWithdraw(object):
    """
    No.48 in-policy reject test
    g2 (asn: 65002)
    g3 (asn: 65003)

    g2's in-policy only accepts routes with origin asn 65002

    r1:192.168.10.0/24

    1.

         r1
          |                                  g1(rs)
          v                             ----------------
    g3 - g2 ->(r1(aspath=[65002]))->  o | -> g4-rib -> | -> r1(aspath=[65002]) --> g4
                                        ----------------

    2. g3 also sends prefix r1 (the prefix from g2 is still the best for the prefix)

    r1   r1
    |     |                                  g1(rs)
    v     v                             ----------------
    g3 - g2 ->(r1(aspath=[65002]))->  o | -> g4-rib -> | -> r1(aspath=[65002]) --> g4
                                        ----------------

    3. g2 withdraws r1, then the path from g3 becomes the best (implicit withdrawal happens).
       Since g2's in-policy only accepts routes with origin asn 2, rs must send withdrawal to g4.

    r1   r1
    |     x                                        g1(rs)
    v                                         ----------------
    g3 - g2 ->(r1(aspath=[65002,65003]))->  x | -> g4-rib -> | -> r1(withdrawal) --> g4
                                              ----------------
    """
    @staticmethod
    def boot(env):
        gobgp_ctn_image_name = env.parser_option.gobgp_image
        log_level = env.parser_option.gobgp_log_level
        g1 = GoBGPContainer(name='g1', asn=65001, router_id='192.168.0.1',
                            ctn_image_name=gobgp_ctn_image_name,
                            log_level=log_level)
        g2 = GoBGPContainer(name='g2', asn=65002, router_id='192.168.0.2',
                            ctn_image_name=gobgp_ctn_image_name,
                            log_level=log_level)
        g3 = GoBGPContainer(name='g3', asn=65003, router_id='192.168.0.3',
                            ctn_image_name=gobgp_ctn_image_name,
                            log_level=log_level)
        g4 = GoBGPContainer(name='g4', asn=65004, router_id='192.168.0.4',
                            ctn_image_name=gobgp_ctn_image_name,
                            log_level=log_level)

        ctns = [g1, g2, g3, g4]
        initial_wait_time = max(ctn.run() for ctn in ctns)
        time.sleep(initial_wait_time)

        for cli in [g2, g4]:
            g1.add_peer(cli, is_rs_client=True)
            cli.add_peer(g1)

        g3.add_peer(g2)
        g2.add_peer(g3)

        env.g1 = g1
        env.g2 = g2
        env.g3 = g3
        env.g4 = g4

    @staticmethod
    def setup(env):
        g1 = env.g1
        g2 = env.g2
        # g3 = env.g3
        g4 = env.g4

        as0 = {'as-path-sets': [{'as-path-set-name': 'as0', 'as-path-list': ['_65002$']}]}

        g1.set_bgp_defined_set(as0)

        st0 = {'name': 'st0',
               'conditions': {'bgp-conditions': {'match-as-path-set': {'as-path-set': 'as0'}}},
               'actions': {'route-disposition': 'accept-route'}}

        policy = {'name': 'policy0',
                  'statements': [st0]}
        g1.add_policy(policy, g2, 'import', 'reject')

        g2.add_route('192.168.0.0/24')

        for c in [g2, g4]:
            g1.wait_for(BGP_FSM_ESTABLISHED, c)

        g2.wait_for(BGP_FSM_ESTABLISHED, g1)

    @staticmethod
    def check(env):
        g1 = env.g1
        # g2 = env.g2
        g4 = env.g4
        wait_for(lambda: len(g1.get_local_rib(g4)) == 1)
        wait_for(lambda: len(g1.get_local_rib(g4)[0]['paths']) == 1)
        wait_for(lambda: len(g4.get_global_rib()) == 1)
        wait_for(lambda: len(g4.get_global_rib()[0]['paths']) == 1)

    @staticmethod
    def setup2(env):
        env.g3.add_route('192.168.0.0/24')

    @staticmethod
    def check2(env):
        g1 = env.g1
        g2 = env.g2
        g4 = env.g4
        wait_for(lambda: len(g2.get_global_rib()) == 1)
        wait_for(lambda: len(g2.get_global_rib()[0]['paths']) == 2)
        wait_for(lambda: len(g1.get_local_rib(g4)) == 1)
        wait_for(lambda: len(g1.get_local_rib(g4)[0]['paths']) == 1)
        wait_for(lambda: len(g1.get_adj_rib_in(g2)) == 1)
        wait_for(lambda: len(g4.get_global_rib()) == 1)
        wait_for(lambda: len(g4.get_global_rib()[0]['paths']) == 1)

    @staticmethod
    def setup3(env):
        env.g2.local('gobgp global rib del 192.168.0.0/24')

    @staticmethod
    def check3(env):
        g1 = env.g1
        g2 = env.g2
        g4 = env.g4
        wait_for(lambda: len(g2.get_global_rib()) == 1)
        wait_for(lambda: len(g2.get_global_rib()[0]['paths']) == 1)
        wait_for(lambda: len(g1.get_adj_rib_in(g2)) == 1)
        wait_for(lambda: len(g1.get_local_rib(g4)) == 0)
        wait_for(lambda: len(g4.get_global_rib()) == 0)

    @staticmethod
    def executor(env):
        lookup_scenario("ImportPolicyRejectImplicitWithdraw").boot(env)
        lookup_scenario("ImportPolicyRejectImplicitWithdraw").setup(env)
        lookup_scenario("ImportPolicyRejectImplicitWithdraw").check(env)
        lookup_scenario("ImportPolicyRejectImplicitWithdraw").setup2(env)
        lookup_scenario("ImportPolicyRejectImplicitWithdraw").check2(env)
        lookup_scenario("ImportPolicyRejectImplicitWithdraw").setup3(env)
        lookup_scenario("ImportPolicyRejectImplicitWithdraw").check3(env)


class TestGoBGPBase(unittest.TestCase):

    wait_per_retry = 5
    retry_limit = 10

    @classmethod
    def setUpClass(cls):
        idx = parser_option.test_index
        base.TEST_PREFIX = parser_option.test_prefix
        cls.parser_option = parser_option
        cls.executors = []
        if idx == 0:
            print('unset test-index. run all test sequential')
            for _, v in list(_SCENARIOS.items()):
                for k, m in inspect.getmembers(v, inspect.isfunction):
                    if k == 'executor':
                        cls.executor = m
                cls.executors.append(cls.executor)
        elif idx not in _SCENARIOS:
            print('invalid test-index. # of scenarios: {0}'.format(len(_SCENARIOS)))
            sys.exit(1)
        else:
            for k, m in inspect.getmembers(_SCENARIOS[idx], inspect.isfunction):
                if k == 'executor':
                    cls.executor = m
            cls.executors.append(cls.executor)

    def test(self):
        for e in self.executors:
            e(self)
            print('[PASS] %s' % e.__qualname__.split('.')[0], file=sys.stderr, flush=True)





# Copyright (C) 2015 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from itertools import combinations
import sys
import time
import unittest

import collections
collections.Callable = collections.abc.Callable


from lib.noseplugin import parser_option

from lib import base
from lib.base import (
    BGP_FSM_ESTABLISHED,
    BGP_ATTR_TYPE_EXTENDED_COMMUNITIES,
    local,
)
from lib.gobgp import GoBGPContainer


def get_mac_mobility_sequence(pattr):
    for ecs in [
            p['value'] for p in pattr
            if 'type' in p and p['type'] == BGP_ATTR_TYPE_EXTENDED_COMMUNITIES]:
        for ec in [e for e in ecs if 'type' in e and e['type'] == 6]:
            if ec['subtype'] == 0:
                if 'sequence' not in ec:
                    return 0
                else:
                    return ec['sequence']
    return -1


class GoBGPTestBase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        gobgp_ctn_image_name = parser_option.gobgp_image
        base.TEST_PREFIX = parser_option.test_prefix

        g1 = GoBGPContainer(name='g1', asn=65000, router_id='192.168.0.1',
                            ctn_image_name=gobgp_ctn_image_name,
                            log_level=parser_option.gobgp_log_level)
        g2 = GoBGPContainer(name='g2', asn=65000, router_id='192.168.0.2',
                            ctn_image_name=gobgp_ctn_image_name,
                            log_level=parser_option.gobgp_log_level)
        ctns = [g1, g2]

        initial_wait_time = max(ctn.run() for ctn in ctns)

        time.sleep(initial_wait_time)
        g1.local("gobgp vrf add vrf1 rd 10:10 rt both 10:10")
        g1.local("gobgp vrf add vrf2 rd 10:20 rt both 10:20")
        g2.local("gobgp vrf add vrf1 rd 20:10 rt both 10:10")
        g2.local("gobgp vrf add vrf2 rd 20:20 rt both 10:20")

        for a, b in combinations(ctns, 2):
            a.add_peer(b, vpn=True, passwd='evpn')
            b.add_peer(a, vpn=True, passwd='evpn')

        cls.g1 = g1
        cls.g2 = g2

    # test each neighbor state is turned establish
    def test_01_neighbor_established(self):
        self.g1.wait_for(expected_state=BGP_FSM_ESTABLISHED, peer=self.g2)

    def test_02_add_evpn_route(self):
        self.g1.local('gobgp global rib add '
                      '-a evpn macadv 11:22:33:44:55:66 10.0.0.1 esi AS 1 1 1 etag 1000 label 1000 '
                      'rd 10:10 rt 10:10')
        grib = self.g1.get_global_rib(rf='evpn')
        self.assertEqual(len(grib), 1)
        dst = grib[0]
        self.assertEqual(len(dst['paths']), 1)
        path = dst['paths'][0]
        self.assertEqual(path['nexthop'], '0.0.0.0')

        interval = 1
        timeout = int(30 / interval)
        done = False
        for _ in range(timeout):
            if done:
                break
            grib = self.g2.get_global_rib(rf='evpn')

            if len(grib) < 1:
                time.sleep(interval)
                continue

            self.assertEqual(len(grib), 1)
            dst = grib[0]
            self.assertEqual(len(dst['paths']), 1)
            path = dst['paths'][0]
            n_addrs = [i[1].split('/')[0] for i in self.g1.ip_addrs]
            self.assertTrue(path['nexthop'] in n_addrs)
            done = True

    def test_03_check_mac_mobility(self):
        self.g2.local('gobgp global rib add '
                      '-a evpn macadv 11:22:33:44:55:66 10.0.0.1 esi AS 2 1 1 etag 1000 label 1000 '
                      'rd 20:10 rt 10:10')

        time.sleep(3)

        grib = self.g1.get_global_rib(rf='evpn')
        self.assertEqual(len(grib), 1)
        dst = grib[0]
        self.assertEqual(len(dst['paths']), 1)
        path = dst['paths'][0]
        n_addrs = [i[1].split('/')[0] for i in self.g2.ip_addrs]
        self.assertTrue(path['nexthop'] in n_addrs)
        self.assertEqual(get_mac_mobility_sequence(path['attrs']), 0)

    def test_04_check_mac_mobility_again(self):
        self.g1.local('gobgp global rib add '
                      '-a evpn macadv 11:22:33:44:55:66 10.0.0.1 esi AS 3 1 1 etag 1000 label 1000 '
                      'rd 10:10 rt 10:10')

        time.sleep(3)

        grib = self.g2.get_global_rib(rf='evpn')
        self.assertEqual(len(grib), 1)
        dst = grib[0]
        self.assertEqual(len(dst['paths']), 1)
        path = dst['paths'][0]
        n_addrs = [i[1].split('/')[0] for i in self.g1.ip_addrs]
        self.assertTrue(path['nexthop'] in n_addrs)
        self.assertEqual(get_mac_mobility_sequence(path['attrs']), 1)

    def test_05_check_mac_mobility_per_mac_vrf(self):
        self.g2.local('gobgp global rib add '
                      '-a evpn macadv 11:22:33:44:55:66 10.0.0.1 esi AS 4 1 1 etag 2000 label 2000 '
                      'rd 20:20 rt 10:20')

        time.sleep(3)

        grib = self.g2.get_global_rib(rf='evpn')
        self.assertEqual(len(grib), 2)
        # first route is from previous tests
        dst = grib[0]
        self.assertEqual(len(dst['paths']), 1)
        path = dst['paths'][0]
        n_addrs = [i[1].split('/')[0] for i in self.g1.ip_addrs]
        self.assertTrue(path['nexthop'] in n_addrs)
        self.assertEqual(get_mac_mobility_sequence(path['attrs']), 1)

        # dump global rib again on other gobgp instance to have our second route have the nexthop
        # filled out. otherwise it'd be 0.0.0.0.
        grib = self.g1.get_global_rib(rf='evpn')
        self.assertEqual(len(grib), 2)
        # second route from this test, in another mac-vrf
        dst = grib[1]
        self.assertEqual(len(dst['paths']), 1)
        path = dst['paths'][0]
        n_addrs = [i[1].split('/')[0] for i in self.g2.ip_addrs]
        self.assertTrue(path['nexthop'] in n_addrs)
        # no mac mobility for this route
        self.assertEqual(get_mac_mobility_sequence(path['attrs']), -1)





# Copyright (C) 2015 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import collections
import json
from itertools import chain
from threading import Thread
import subprocess
import os

import netaddr
import toml
import yaml

from lib.base import (
    community_str,
    wait_for_completion,
    BGPContainer,
    CmdBuffer,
    BGP_ATTR_TYPE_AS_PATH,
    BGP_ATTR_TYPE_NEXT_HOP,
    BGP_ATTR_TYPE_MULTI_EXIT_DISC,
    BGP_ATTR_TYPE_LOCAL_PREF,
    BGP_ATTR_TYPE_COMMUNITIES,
    BGP_ATTR_TYPE_MP_REACH_NLRI,
    GRACEFUL_RESTART_TIME,
    LONG_LIVED_GRACEFUL_RESTART_TIME,
    BGP_FSM_IDLE,
    BGP_FSM_ACTIVE,
    BGP_FSM_ESTABLISHED,
    yellow,
    indent,
    local,
)


def extract_path_attribute(path, typ):
    for a in path['attrs']:
        if a['type'] == typ:
            return a
    return None


class GoBGPContainer(BGPContainer):

    SHARED_VOLUME = '/root/shared_volume'
    QUAGGA_VOLUME = '/etc/quagga'

    def __init__(self, name, asn, router_id, ctn_image_name='osrg/gobgp',
                 log_level='debug', zebra=False, config_format='toml',
                 zapi_version=2, bgp_config=None, ospfd_config=None,
                 zebra_multipath_enabled=False):
        super(GoBGPContainer, self).__init__(name, asn, router_id,
                                             ctn_image_name)
        self.shared_volumes.append((self.config_dir, self.SHARED_VOLUME))
        self.quagga_config_dir = '{0}/quagga'.format(self.config_dir)
        self.shared_volumes.append((self.quagga_config_dir, self.QUAGGA_VOLUME))

        self.log_level = log_level
        self.prefix_set = None
        self.neighbor_set = None
        self.bgp_set = None
        self.statements = None
        self.default_policy = None
        self.zebra = zebra
        self.zapi_version = zapi_version
        self.zebra_multipath_enabled = zebra_multipath_enabled
        self.config_format = config_format

        # bgp_config is equivalent to config.BgpConfigSet structure
        # Example:
        # bgp_config = {
        #     'global': {
        #         'confederation': {
        #             'config': {
        #                 'identifier': 10,
        #                 'member-as-list': [65001],
        #             }
        #         },
        #     },
        # }
        self.bgp_config = bgp_config or {}

        # To start OSPFd in GoBGP container, specify 'ospfd_config' as a dict
        # type value.
        # Example:
        # ospfd_config = {
        #     'redistributes': [
        #         'connected',
        #     ],
        #     'networks': {
        #         '192.168.1.0/24': '0.0.0.0',  # <network>: <area>
        #     },
        # }
        self.ospfd_config = ospfd_config or {}

    def _start_gobgp(self, graceful_restart=False):
        c = CmdBuffer()
        c << '#!/bin/sh'
        c << '/go/bin/gobgpd -f {0}/gobgpd.conf -l {1} -p {2} -t {3} > ' \
             '{0}/gobgpd.log 2>&1'.format(self.SHARED_VOLUME, self.log_level, '-r' if graceful_restart else '', self.config_format)
        cmd = 'echo "{0:s}" > {1}/start.sh'.format(str(c), self.config_dir)
        local(cmd, capture=True)
        cmd = "chmod 755 {0}/start.sh".format(self.config_dir)
        local(cmd, capture=True)
        self.local("{0}/start.sh".format(self.SHARED_VOLUME), detach=True)

    def start_gobgp(self, graceful_restart=False):
        if self._is_running():
            raise RuntimeError('GoBGP is already running')
        self._start_gobgp(graceful_restart=graceful_restart)
        self._wait_for_boot()

    def stop_gobgp(self):
        self.local("pkill -KILL gobgpd")

    def _start_zebra(self):
        if self.zapi_version == 2:
            daemon_bin = '/usr/lib/quagga/zebra'
        else:
            daemon_bin = 'zebra'
        cmd = '{0} -f {1}/zebra.conf'.format(daemon_bin, self.QUAGGA_VOLUME)
        self.local(cmd, detach=True)

    def _start_ospfd(self):
        if self.zapi_version == 2:
            daemon_bin = '/usr/lib/quagga/ospfd'
        else:
            daemon_bin = 'ospfd'
        cmd = '{0} -f {1}/ospfd.conf'.format(daemon_bin, self.QUAGGA_VOLUME)
        self.local(cmd, detach=True)

    def _get_enabled_quagga_daemons(self):
        daemons = []
        if self.zebra:
            daemons.append('zebra')
            if self.ospfd_config:
                daemons.append('ospfd')
        return daemons

    def _is_running(self):
        return self.local('gobgp global'
                          ' > /dev/null 2>&1; echo $?', capture=True) == '0'

    def _wait_for_boot(self):
        for daemon in self._get_enabled_quagga_daemons():
            def _f_quagga():
                ret = self.local("vtysh -d {0} -c 'show run' > /dev/null 2>&1; echo $?".format(daemon), capture=True)
                return ret == '0'

            wait_for_completion(_f_quagga)

        wait_for_completion(self._is_running)

    def run(self):
        super(GoBGPContainer, self).run()
        if self.zebra:
            self._start_zebra()
            if self.ospfd_config:
                self._start_ospfd()
        self._start_gobgp()
        self._wait_for_boot()
        return self.WAIT_FOR_BOOT

    @staticmethod
    def _get_as_path(path):
        asps = (p['as_paths'] for p in path['attrs']
                if p['type'] == BGP_ATTR_TYPE_AS_PATH and 'as_paths' in p and p['as_paths'] is not None)
        asps = chain.from_iterable(asps)
        asns = (asp['asns'] for asp in asps)
        return list(chain.from_iterable(asns))

    @staticmethod
    def _get_nexthop(path):
        for p in path['attrs']:
            if p['type'] == BGP_ATTR_TYPE_NEXT_HOP or p['type'] == BGP_ATTR_TYPE_MP_REACH_NLRI:
                return p['nexthop']

    @staticmethod
    def _get_local_pref(path):
        for p in path['attrs']:
            if p['type'] == BGP_ATTR_TYPE_LOCAL_PREF:
                return p['value']
        return None

    @staticmethod
    def _get_med(path):
        for p in path['attrs']:
            if p['type'] == BGP_ATTR_TYPE_MULTI_EXIT_DISC:
                return p['metric']
        return None

    @staticmethod
    def _get_community(path):
        for p in path['attrs']:
            if p['type'] == BGP_ATTR_TYPE_COMMUNITIES:
                return [community_str(c) for c in p['communities']]
        return None

    def _get_rib(self, dests_dict):
        dests = []
        for k, v in list(dests_dict.items()):
            for p in v:
                p["nexthop"] = self._get_nexthop(p)
                p["aspath"] = self._get_as_path(p)
                p["local-pref"] = self._get_local_pref(p)
                p["community"] = self._get_community(p)
                p["med"] = self._get_med(p)
                p["prefix"] = k
                path_id = p.get("id", None)
                if path_id:
                    p["identifier"] = p["id"]
            dests.append({'paths': v, 'prefix': k})
        return dests

    def _trigger_peer_cmd(self, cmd, peer):
        peer_addr = self.peer_name(peer)
        cmd = 'gobgp neighbor {0} {1}'.format(peer_addr, cmd)
        self.local(cmd)

    def disable_peer(self, peer):
        self._trigger_peer_cmd('disable', peer)

    def enable_peer(self, peer):
        self._trigger_peer_cmd('enable', peer)

    def reset(self, peer):
        self._trigger_peer_cmd('reset', peer)

    def softreset(self, peer, rf='ipv4', type='in'):
        self._trigger_peer_cmd('softreset{0} -a {1}'.format(type, rf), peer)

    def get_local_rib(self, peer, prefix='', rf='ipv4'):
        peer_addr = self.peer_name(peer)
        cmd = 'gobgp -j neighbor {0} local {1} -a {2}'.format(peer_addr, prefix, rf)
        output = self.local(cmd, capture=True)
        return self._get_rib(json.loads(output))

    def get_global_rib(self, prefix='', rf='ipv4'):
        cmd = 'gobgp -j global rib {0} -a {1}'.format(prefix, rf)
        output = self.local(cmd, capture=True)
        return self._get_rib(json.loads(output))

    def monitor_global_rib(self, queue, rf='ipv4'):
        host = self.ip_addrs[0][1].split('/')[0]

        if not os.path.exists('{0}/gobgp'.format(self.config_dir)):
            self.local('cp /go/bin/gobgp {0}/'.format(self.SHARED_VOLUME))

        args = '{0}/gobgp -u {1} -j monitor global rib -a {2}'.format(self.config_dir, host, rf).split(' ')

        def monitor():
            process = subprocess.Popen(args, stdout=subprocess.PIPE)
            for line in iter(process.stdout.readline, ''):
                p = json.loads(line)[0]
                p["nexthop"] = self._get_nexthop(p)
                p["aspath"] = self._get_as_path(p)
                p["local-pref"] = self._get_local_pref(p)
                p["med"] = self._get_med(p)
                queue.put(p)

        t = Thread(target=monitor)
        t.daemon = True
        t.start()

    def _get_adj_rib(self, adj_type, peer, prefix='', rf='ipv4', add_path_enabled=False):
        peer_addr = self.peer_name(peer)
        cmd = 'gobgp neighbor {0} adj-{1} {2} -a {3} -j'.format(peer_addr,
                                                                adj_type,
                                                                prefix, rf)
        output = self.local(cmd, capture=True)
        if add_path_enabled:
            return self._get_rib(json.loads(output))

        ret = [p[0] for p in json.loads(output).values()]
        for p in ret:
            p["nexthop"] = self._get_nexthop(p)
            p["aspath"] = self._get_as_path(p)
            p["prefix"] = p['nlri']['prefix']
            p["local-pref"] = self._get_local_pref(p)
            p["med"] = self._get_med(p)
        return ret

    def get_adj_rib_in(self, peer, prefix='', rf='ipv4', add_path_enabled=False):
        return self._get_adj_rib('in', peer, prefix, rf, add_path_enabled)

    def get_adj_rib_out(self, peer, prefix='', rf='ipv4', add_path_enabled=False):
        return self._get_adj_rib('out', peer, prefix, rf, add_path_enabled)

    def get_neighbor(self, peer):
        cmd = 'gobgp -j neighbor {0}'.format(self.peer_name(peer))
        return json.loads(self.local(cmd, capture=True))

    def get_neighbor_state(self, peer):
        s = self.get_neighbor(peer)['state']['session_state']
        if s == 1:
            return BGP_FSM_IDLE
        elif s == 3:
            return BGP_FSM_ACTIVE
        elif s == 6:
            return BGP_FSM_ESTABLISHED
        return "unknown"

    def clear_policy(self):
        self.policies = {}
        for info in self.peers.values():
            info['policies'] = {}
        self.prefix_set = []
        self.neighbor_set = []
        self.statements = []

    def set_prefix_set(self, ps):
        if not isinstance(ps, list):
            ps = [ps]
        self.prefix_set = ps

    def add_prefix_set(self, ps):
        if self.prefix_set is None:
            self.prefix_set = []
        self.prefix_set.append(ps)

    def set_neighbor_set(self, ns):
        if not isinstance(ns, list):
            ns = [ns]
        self.neighbor_set = ns

    def add_neighbor_set(self, ns):
        if self.neighbor_set is None:
            self.neighbor_set = []
        self.neighbor_set.append(ns)

    def set_bgp_defined_set(self, bs):
        self.bgp_set = bs

    def create_config(self):
        self._create_config_bgp()
        if self.zebra:
            local('mkdir -p {0}'.format(self.quagga_config_dir))
            local('chmod 777 {0}'.format(self.quagga_config_dir))
            self._create_config_zebra()
            if self.ospfd_config:
                self._create_config_ospfd()

    def _merge_dict(self, dct, merge_dct):
        for k, v in merge_dct.items():
            if (k in dct and isinstance(dct[k], dict)
                    and isinstance(merge_dct[k], collections.abc.Mapping)):
                self._merge_dict(dct[k], merge_dct[k])
            else:
                dct[k] = merge_dct[k]

    def _create_config_bgp(self):
        config = {
            'global': {
                'config': {
                    'as': self.asn,
                    'router-id': self.router_id,
                },
                'route-selection-options': {
                    'config': {
                        'external-compare-router-id': True,
                    },
                },
            },
            'neighbors': [],
        }

        self._merge_dict(config, self.bgp_config)

        if self.zebra and self.zapi_version == 2:
            config['global']['use-multiple-paths'] = {'config': {'enabled': True}}
        else:
            config['global']['use-multiple-paths'] = {'config': {'enabled': self.zebra_multipath_enabled}}

        for peer, info in self.peers.items():
            afi_safi_list = []
            if info['interface'] != '':
                afi_safi_list.append({'config': {'afi-safi-name': 'ipv4-unicast'}})
                afi_safi_list.append({'config': {'afi-safi-name': 'ipv6-unicast'}})
            else:
                version = netaddr.IPNetwork(info['neigh_addr']).version
                if version == 4:
                    afi_safi_list.append({'config': {'afi-safi-name': 'ipv4-unicast'}})
                elif version == 6:
                    afi_safi_list.append({'config': {'afi-safi-name': 'ipv6-unicast'}})
                else:
                    Exception('invalid ip address version. {0}'.format(version))

            if info['vpn']:
                afi_safi_list.append({'config': {'afi-safi-name': 'l3vpn-ipv4-unicast'}})
                afi_safi_list.append({'config': {'afi-safi-name': 'l3vpn-ipv6-unicast'}})
                afi_safi_list.append({'config': {'afi-safi-name': 'l2vpn-evpn'}})
                afi_safi_list.append({'config': {'afi-safi-name': 'rtc'}, 'route-target-membership': {'config': {'deferral-time': 10}}})

            if info['flowspec']:
                afi_safi_list.append({'config': {'afi-safi-name': 'ipv4-flowspec'}})
                afi_safi_list.append({'config': {'afi-safi-name': 'l3vpn-ipv4-flowspec'}})
                afi_safi_list.append({'config': {'afi-safi-name': 'ipv6-flowspec'}})
                afi_safi_list.append({'config': {'afi-safi-name': 'l3vpn-ipv6-flowspec'}})

            if info['mup']:
                afi_safi_list.append({'config': {'afi-safi-name': 'ipv4-mup'}})
                afi_safi_list.append({'config': {'afi-safi-name': 'ipv6-mup'}})

            neigh_addr = None
            interface = None
            peer_as = None
            if info['interface'] == '':
                neigh_addr = info['neigh_addr'].split('/')[0]
                peer_as = info['remote_as']
            else:
                interface = info['interface']
            n = {
                'config': {
                    'neighbor-address': neigh_addr,
                    'neighbor-interface': interface,
                    'peer-as': peer_as,
                    'auth-password': info['passwd'],
                    'vrf': info['vrf'],
                    'remove-private-as': info['remove_private_as'],
                },
                'afi-safis': afi_safi_list,
                'timers': {
                    'config': {
                        'connect-retry': 10,
                    },
                },
                'transport': {
                    'config': {},
                },
            }

            n['as-path-options'] = {'config': {}}
            if info['allow_as_in'] > 0:
                n['as-path-options']['config']['allow-own-as'] = info['allow_as_in']
            if info['replace_peer_as']:
                n['as-path-options']['config']['replace-peer-as'] = info['replace_peer_as']

            if ':' in info['local_addr']:
                n['transport']['config']['local-address'] = info['local_addr'].split('/')[0].split('%')[0]

            if info['passive']:
                n['transport']['config']['passive-mode'] = True
            if info['bind_interface']:
                n['transport']['config']['bind-interface'] = info['bind_interface']

            if info['is_rs_client']:
                n['route-server'] = {'config': {'route-server-client': True}}

            if info['local_as']:
                n['config']['local-as'] = info['local_as']

            if info['prefix_limit']:
                for v in afi_safi_list:
                    v['prefix-limit'] = {'config': {'max-prefixes': info['prefix_limit'], 'shutdown-threshold-pct': 80}}

            if info['graceful_restart'] is not None:
                n['graceful-restart'] = {'config': {'enabled': True, 'restart-time': GRACEFUL_RESTART_TIME}}
                for afi_safi in afi_safi_list:
                    afi_safi['mp-graceful-restart'] = {'config': {'enabled': True}}

                if info['llgr'] is not None:
                    n['graceful-restart']['config']['restart-time'] = 1
                    n['graceful-restart']['config']['long-lived-enabled'] = True
                    for afi_safi in afi_safi_list:
                        afi_safi['long-lived-graceful-restart'] = {'config': {'enabled': True, 'restart-time': LONG_LIVED_GRACEFUL_RESTART_TIME}}

            if info['is_rr_client']:
                cluster_id = self.router_id
                if 'cluster_id' in info and info['cluster_id'] is not None:
                    cluster_id = info['cluster_id']
                n['route-reflector'] = {'config': {'route-reflector-client': True,
                                                   'route-reflector-cluster-id': cluster_id}}

            if info["addpath"] > 0:
                n["add-paths"] = {
                    "config": {"receive": True, "send-max": info["addpath"]}
                }

            if len(info.get('default-policy', [])) + len(info.get('policies', [])) > 0:
                n['apply-policy'] = {'config': {}}

            for typ, p in info.get('policies', {}).items():
                n['apply-policy']['config']['{0}-policy-list'.format(typ)] = [p['name']]

            def _f(v):
                if v == 'reject':
                    return 'reject-route'
                elif v == 'accept':
                    return 'accept-route'
                raise Exception('invalid default policy type {0}'.format(v))

            for typ, d in info.get('default-policy', {}).items():
                n['apply-policy']['config']['default-{0}-policy'.format(typ)] = _f(d)

            if info['treat_as_withdraw']:
                n['error-handling'] = {'config': {'treat-as-withdraw': True}}

            config['neighbors'].append(n)

        config['defined-sets'] = {}
        if self.prefix_set:
            config['defined-sets']['prefix-sets'] = self.prefix_set

        if self.neighbor_set:
            config['defined-sets']['neighbor-sets'] = self.neighbor_set

        if self.bgp_set:
            config['defined-sets']['bgp-defined-sets'] = self.bgp_set

        policy_list = []
        for p in self.policies.values():
            policy = {'name': p['name']}
            if 'statements' in p:
                policy['statements'] = p['statements']
            policy_list.append(policy)

        if len(policy_list) > 0:
            config['policy-definitions'] = policy_list

        if self.zebra:
            config['zebra'] = {'config': {'enabled': True,
                                          'redistribute-route-type-list': ['connect'],
                                          'version': self.zapi_version}}

        with open('{0}/gobgpd.conf'.format(self.config_dir), 'w') as f:
            print(yellow('[{0}\'s new gobgpd.conf]'.format(self.name)))
            if self.config_format == "toml":
                raw = toml.dumps(config)
            elif self.config_format == "yaml":
                raw = yaml.dump(config)
            elif self.config_format == "json":
                raw = json.dumps(config)
            else:
                raise Exception('invalid config_format {0}'.format(self.config_format))
            raw = raw.strip()
            print(yellow(indent(raw)))
            f.write(raw)

    def _create_config_zebra(self):
        c = CmdBuffer()
        c << 'hostname zebra'
        c << 'password zebra'
        c << 'log file {0}/zebra.log'.format(self.QUAGGA_VOLUME)
        c << 'debug zebra packet'
        c << 'debug zebra kernel'
        c << 'debug zebra rib'
        c << 'ipv6 forwarding'
        c << ''

        with open('{0}/zebra.conf'.format(self.quagga_config_dir), 'w') as f:
            print(yellow('[{0}\'s new zebra.conf]'.format(self.name)))
            c = str(c).strip()
            print(yellow(indent(c)))
            f.writelines(c)

    def _create_config_ospfd(self):
        c = CmdBuffer()
        c << 'hostname ospfd'
        c << 'password zebra'
        c << 'router ospf'
        for redistribute in self.ospfd_config.get('redistributes', []):
            c << ' redistribute {0}'.format(redistribute)
        for network, area in list(self.ospfd_config.get('networks', {}).items()):
            c << ' network {0} area {1}'.format(network, area)
        c << 'log file {0}/ospfd.log'.format(self.QUAGGA_VOLUME)
        c << ''

        with open('{0}/ospfd.conf'.format(self.quagga_config_dir), 'w') as f:
            print(yellow('[{0}\'s new ospfd.conf]'.format(self.name)))
            print(yellow(indent(str(c))))
            f.writelines(str(c))

    def reload_config(self):
        for daemon in self._get_enabled_quagga_daemons():
            self.local('pkill -SIGHUP {0}'.format(daemon), capture=True)
        self.local('pkill -SIGHUP gobgpd', capture=True)
        self._wait_for_boot()

    def add_route(self, route, rf='ipv4', attribute=None, aspath=None,
                  community=None, med=None, extendedcommunity=None,
                  nexthop=None, matchs=None, thens=None,
                  local_pref=None, identifier=None, reload_config=False):
        if not self._is_running():
            raise RuntimeError('GoBGP is not yet running')

        self.routes.setdefault(route, [])
        path = {
            'prefix': route,
            'rf': rf,
            'attr': attribute,
            'next-hop': nexthop,
            'as-path': aspath,
            'community': community,
            'med': med,
            'local-pref': local_pref,
            'extended-community': extendedcommunity,
            'identifier': identifier,
            'matchs': matchs,
            'thens': thens,
        }

        c = CmdBuffer(' ')
        c << 'gobgp global rib -a {0} add'.format(rf)
        if rf in ('ipv4', 'ipv6'):
            c << route
            if path['identifier']:
                c << 'identifier {0}'.format(path['identifier'])
            if path['next-hop']:
                c << 'nexthop {0}'.format(path['next-hop'])
            if path['local-pref']:
                c << 'local-pref {0}'.format(path['local-pref'])
            if path['med']:
                c << 'med {0}'.format(path['med'])
            if path['community']:
                comm = str(path['community'])
                if isinstance(path['community'], (list, tuple)):
                    comm = ','.join(path['community'])
                c << 'community {0}'.format(comm)
        elif rf.endswith('-flowspec'):
            c << 'match {0}'.format(' '.join(path['matchs']))
            c << 'then {0}'.format(' '.join(path['thens']))
        else:
            raise Exception('unsupported address family: {0}'.format(rf))
        self.local(str(c), capture=True)

        self.routes[route].append(path)

    def del_route(self, route, identifier=None, reload_config=True):
        if not self._is_running():
            raise RuntimeError('GoBGP is not yet running')

        if route not in self.routes:
            return

        new_paths = []
        for path in self.routes[route]:
            if path['identifier'] != identifier:
                new_paths.append(path)
            else:
                r = CmdBuffer(' ')
                r << 'gobgp global -a {0}'.format(path['rf'])
                prefix = path['prefix']
                if path['rf'].endswith('-flowspec'):
                    prefix = 'match {0}'.format(' '.join(path['matchs']))
                r << 'rib del {0}'.format(prefix)
                if identifier:
                    r << 'identifier {0}'.format(identifier)
                cmd = str(r)
                self.local(cmd, capture=True)
        self.routes[route] = new_paths
        # no need to reload config


class RawGoBGPContainer(GoBGPContainer):
    def __init__(self, name, config, ctn_image_name='osrg/gobgp',
                 log_level='debug', zebra=False, config_format='yaml'):
        if config_format == "toml":
            d = toml.loads(config)
        elif config_format == "yaml":
            d = yaml.load(config)
        elif config_format == "json":
            d = json.loads(config)
        else:
            raise Exception('invalid config format {0}'.format(config_format))
        asn = d['global']['config']['as']
        router_id = d['global']['config']['router-id']
        self.config = config
        super(RawGoBGPContainer, self).__init__(name, asn, router_id,
                                                ctn_image_name, log_level,
                                                zebra, config_format)

    def create_config(self):
        with open('{0}/gobgpd.conf'.format(self.config_dir), 'w') as f:
            print(yellow('[{0}\'s new gobgpd.conf]'.format(self.name)))
            print(yellow(indent(self.config)))
            f.write(self.config)
